from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class Browser:
    """
    This class represents the live browser data base.
    """

    def add_filter_type_listener(self, listener: Callable) -> None:
        """
        add_filter_type_listener( (Browser)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "filter_type" has changed.

            C++ signature :
                void add_filter_type_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def add_full_refresh_listener(self, listener: Callable) -> None:
        """
        add_full_refresh_listener( (Browser)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "full_refresh" has changed.

            C++ signature :
                void add_full_refresh_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def add_hotswap_target_listener(self, listener: Callable) -> None:
        """
        add_hotswap_target_listener( (Browser)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "hotswap_target" has changed.

            C++ signature :
                void add_hotswap_target_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def filter_type_has_listener(self, listener: Callable) -> bool:
        """
        filter_type_has_listener( (Browser)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "filter_type".

            C++ signature :
                bool filter_type_has_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def full_refresh_has_listener(self, listener: Callable) -> bool:
        """
        full_refresh_has_listener( (Browser)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "full_refresh".

            C++ signature :
                bool full_refresh_has_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def hotswap_target_has_listener(self, listener: Callable) -> bool:
        """
        hotswap_target_has_listener( (Browser)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "hotswap_target".

            C++ signature :
                bool hotswap_target_has_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def load_item(self, arg2: object) -> None:
        """
        load_item( (Browser)arg1, (BrowserItem)arg2) -> None :
            Loads the provided browser item.

            C++ signature :
                void load_item(TPyHandle<ABrowserDelegate>,NPythonBrowser::TPythonBrowserItem)
        """

    def preview_item(self, arg2: object) -> None:
        """
        preview_item( (Browser)arg1, (BrowserItem)arg2) -> None :
            Previews the provided browser item.

            C++ signature :
                void preview_item(TPyHandle<ABrowserDelegate>,NPythonBrowser::TPythonBrowserItem)
        """

    def relation_to_hotswap_target(self, arg2: object) -> Relation:
        """
        relation_to_hotswap_target( (Browser)arg1, (BrowserItem)arg2) -> Relation :
            Returns the relation between the given browser item and the current hotswap target

            C++ signature :
                ableton::live_library::Relation relation_to_hotswap_target(TPyHandle<ABrowserDelegate>,NPythonBrowser::TPythonBrowserItem)
        """

    def remove_filter_type_listener(self, listener: Callable) -> None:
        """
        remove_filter_type_listener( (Browser)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "filter_type".

            C++ signature :
                void remove_filter_type_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def remove_full_refresh_listener(self, listener: Callable) -> None:
        """
        remove_full_refresh_listener( (Browser)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "full_refresh".

            C++ signature :
                void remove_full_refresh_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def remove_hotswap_target_listener(self, listener: Callable) -> None:
        """
        remove_hotswap_target_listener( (Browser)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "hotswap_target".

            C++ signature :
                void remove_hotswap_target_listener(TPyHandle<ABrowserDelegate>,boost::python::api::object)
        """

    def stop_preview(self) -> None:
        """
        stop_preview( (Browser)arg1) -> None :
            Stop the current preview.

            C++ signature :
                void stop_preview(TPyHandle<ABrowserDelegate>)
        """

    @property
    def audio_effects(self) -> Any:
        """
        Returns a browser item with access to all the Audio Effects content.
        """

    @property
    def clips(self) -> list[Clip]:
        """
        Returns a browser item with access to all the Clips content.
        """

    @property
    def colors(self) -> Any:
        """
        Returns a list of browser items containing the configured colors.
        """

    @property
    def current_project(self) -> Any:
        """
        Returns a browser item with access to all the Current Project content.
        """

    @property
    def drums(self) -> Any:
        """
        Returns a browser item with access to all the Drums content.
        """

    @property
    def filter_type(self) -> Any:
        """
        Bang triggered when the hotswap target has changed.
        """

    @property
    def hotswap_target(self) -> Any:
        """
        Bang triggered when the hotswap target has changed.
        """

    @property
    def instruments(self) -> Any:
        """
        Returns a browser item with access to all the Instruments content.
        """

    @property
    def legacy_libraries(self) -> Any:
        """
        Returns a list of browser items containing the installed legacy libraries. The list is always empty as legacy library handling has been removed.
        """

    @property
    def max_for_live(self) -> Any:
        """
        Returns a browser item with access to all the Max For Live content.
        """

    @property
    def midi_effects(self) -> Any:
        """
        Returns a browser item with access to all the Midi Effects content.
        """

    @property
    def packs(self) -> Any:
        """
        Returns a browser item with access to all the Packs content.
        """

    @property
    def plugins(self) -> Any:
        """
        Returns a browser item with access to all the Plugins content.
        """

    @property
    def samples(self) -> Any:
        """
        Returns a browser item with access to all the Samples content.
        """

    @property
    def sounds(self) -> Any:
        """
        Returns a browser item with access to all the Sounds content.
        """

    @property
    def user_folders(self) -> Any:
        """
        Returns a list of browser items containing all the user folders.
        """

    @property
    def user_library(self) -> Any:
        """
        Returns a browser item with access to all the User Library content.
        """


class BrowserItem:
    """
    This class represents an item of the browser hierarchy.
    """

    @property
    def children(self) -> Any:
        """
        Const access to the descendants of this browser item.
        """

    @property
    def is_device(self) -> bool:
        """
        Indicates if the browser item represents a device.
        """

    @property
    def is_folder(self) -> bool:
        """
        Indicates if the browser item represents folder.
        """

    @property
    def is_loadable(self) -> bool:
        """
        True if item can be loaded via the Browser's 'load_item' method.
        """

    @property
    def is_selected(self) -> bool:
        """
        True if the item is ancestor of or the actual selection.
        """

    @property
    def iter_children(self) -> Any:
        """
        Const iterable access to the descendants of this browser item.
        """

    @property
    def name(self) -> str:
        """
        Const access to the canonical display name of this browser item.
        """

    @property
    def source(self) -> Any:
        """
        Specifies where does item come from -- i.e. Live pack, user library...
        """

    @property
    def uri(self) -> Any:
        """
        The uri describes a unique identifier for a browser item.
        """


class BrowserItemIterator:
    """
    This class iterates over children of another BrowserItem.
    """


class BrowserItemVector:
    """
    A container for returning browser items from Live.
    """

    def append(self, arg2: object) -> None:
        """
        append( (BrowserItemVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<NPythonBrowser::TPythonBrowserItem, std::__1::allocator<NPythonBrowser::TPythonBrowserItem>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (BrowserItemVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<NPythonBrowser::TPythonBrowserItem, std::__1::allocator<NPythonBrowser::TPythonBrowserItem>> {lvalue},boost::python::api::object)
        """


class FilterType:
    """
    Class that represent an enumeration of values for FilterType
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    disabled = 0
    hotswap_off = 1
    instrument_hotswap = 2
    audio_effect_hotswap = 3
    midi_effect_hotswap = 4
    drum_pad_hotswap = 5
    midi_track_devices = 6
    samples = 7
    count = 8


class Relation:
    """
    Class that represent an enumeration of values for Relation
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    ancestor = 0
    equal = 1
    descendant = 2
    none = 3



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class DeviceIO:
    """
    This class represents a specific input or output bus of a device.
    """

    def add_available_routing_channels_listener(self, listener: Callable) -> None:
        """
        add_available_routing_channels_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "available_routing_channels" has changed.

            C++ signature :
                void add_available_routing_channels_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def add_available_routing_types_listener(self, listener: Callable) -> None:
        """
        add_available_routing_types_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "available_routing_types" has changed.

            C++ signature :
                void add_available_routing_types_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def add_routing_channel_listener(self, listener: Callable) -> None:
        """
        add_routing_channel_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "routing_channel" has changed.

            C++ signature :
                void add_routing_channel_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def add_routing_type_listener(self, listener: Callable) -> None:
        """
        add_routing_type_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "routing_type" has changed.

            C++ signature :
                void add_routing_type_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def available_routing_channels_has_listener(self, listener: Callable) -> bool:
        """
        available_routing_channels_has_listener( (DeviceIO)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "available_routing_channels".

            C++ signature :
                bool available_routing_channels_has_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def available_routing_types_has_listener(self, listener: Callable) -> bool:
        """
        available_routing_types_has_listener( (DeviceIO)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "available_routing_types".

            C++ signature :
                bool available_routing_types_has_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def remove_available_routing_channels_listener(self, listener: Callable) -> None:
        """
        remove_available_routing_channels_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "available_routing_channels".

            C++ signature :
                void remove_available_routing_channels_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def remove_available_routing_types_listener(self, listener: Callable) -> None:
        """
        remove_available_routing_types_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "available_routing_types".

            C++ signature :
                void remove_available_routing_types_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def remove_routing_channel_listener(self, listener: Callable) -> None:
        """
        remove_routing_channel_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "routing_channel".

            C++ signature :
                void remove_routing_channel_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def remove_routing_type_listener(self, listener: Callable) -> None:
        """
        remove_routing_type_listener( (DeviceIO)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "routing_type".

            C++ signature :
                void remove_routing_type_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def routing_channel_has_listener(self, listener: Callable) -> bool:
        """
        routing_channel_has_listener( (DeviceIO)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "routing_channel".

            C++ signature :
                bool routing_channel_has_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    def routing_type_has_listener(self, listener: Callable) -> bool:
        """
        routing_type_has_listener( (DeviceIO)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "routing_type".

            C++ signature :
                bool routing_type_has_listener(TPyHandle<AMxDRoutable>,boost::python::api::object)
        """

    @property
    def available_routing_channels(self) -> Any:
        """
        Return a list of channels for this IO endpoint.
        """

    @property
    def available_routing_types(self) -> Any:
        """
        Return a list of available routing types for this IO endpoint.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the device IO.
        """

    @property
    def default_external_routing_channel_is_none(self) -> Any:
        """
        Get and set whether the default routing channel for External routing types is none.
        """

    @property
    def routing_channel(self) -> Any:
        """
        Get and set the current routing channel.
        Raises ValueError if the channel isn't one of the current values in
        available_routing_channels.
        """

    @property
    def routing_type(self) -> Any:
        """
        Get and set the current routing type.
        Raises ValueError if the type isn't one of the current values in
        available_routing_types.
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/step_sequence.py
# Compiled at: 2023-12-18 13:26:15
# Size of source mod 2**32: 8240 bytes
from Live.Song import Quantization

from ...base import EventObject, depends, listenable_property, listens
from ...live import get_bar_length, liveobj_valid, prepare_new_clip_slot, song
from .. import Component
from . import LoopSelectorComponent, NoteEditorComponent, NoteEditorPaginator, PlayheadComponent


def create_sequencer_clip(track, length=None):
    length = length or get_bar_length()
    slot = prepare_new_clip_slot(track)
    slot.create_clip(length)
    slot.fire(force_legato=True, launch_quantization=(Quantization.q_no_q))
    song().view.detail_clip = slot.clip
    return slot.clip


class SequencerClip(EventObject):
    @depends(target_track=None)
    def __init__(self, target_track=None, *a, **k):
        (super().__init__)(*a, **k)
        self._clip = None
        self._target_track = target_track
        self._SequencerClip__on_target_clip_changed.subject = target_track
        self._SequencerClip__on_target_clip_changed()

    @listenable_property
    def clip(self):
        return self._clip

    @listenable_property
    def length(self):
        if liveobj_valid(self._clip):
            return self._clip.loop_end - self._clip.loop_start
        return 0

    @listenable_property
    def bar_length(self):
        return get_bar_length(clip=(self._clip))

    @property
    def num_bars(self):
        return self.length / self.bar_length

    def create_clip(self, length=None):
        if self._target_track.target_track.has_midi_input:
            self._clip = create_sequencer_clip((self._target_track.target_track), length=length)
            return self._clip
        return None

    @listens("target_clip")
    def __on_target_clip_changed(self):
        clip = self._target_track.target_clip
        self._clip = clip if (liveobj_valid(clip)) and (clip.is_midi_clip) else None
        self._SequencerClip__on_looping_changed.subject = self._clip
        self._SequencerClip__on_loop_start_changed.subject = self._clip
        self._SequencerClip__on_loop_end_changed.subject = self._clip
        self._SequencerClip__on_signature_numerator_changed.subject = self._clip
        self._SequencerClip__on_signature_denominator_changed.subject = self._clip
        self.notify_clip()
        self.notify_length()
        self.notify_bar_length()

    @listens("looping")
    def __on_looping_changed(self):
        self.notify_length()

    @listens("loop_start")
    def __on_loop_start_changed(self):
        self.notify_length()

    @listens("loop_end")
    def __on_loop_end_changed(self):
        self.notify_length()

    @listens("signature_numerator")
    def __on_signature_numerator_changed(self):
        self.notify_length()
        self.notify_bar_length()

    @listens("signature_denominator")
    def __on_signature_denominator_changed(self):
        self.notify_length()
        self.notify_bar_length()


class StepSequenceComponent(Component):
    @depends(grid_resolution=None)
    def __init__(
        self,
        name="Step_Sequence",
        grid_resolution=None,
        note_editor_component_type=None,
        note_editor_paginator_type=None,
        loop_selector_component_type=None,
        playhead_component_type=None,
        playhead_notes=None,
        playhead_triplet_notes=None,
        playhead_channels=None,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._grid_resolution = self.add_children(grid_resolution)
        note_editor_component_type = note_editor_component_type or NoteEditorComponent
        self._note_editor = note_editor_component_type(grid_resolution=(self._grid_resolution), parent=self)
        note_editor_paginator_type = note_editor_paginator_type or NoteEditorPaginator
        paginator = note_editor_paginator_type(note_editor=(self._note_editor), parent=self)
        loop_selector_component_type = loop_selector_component_type or LoopSelectorComponent
        self._loop_selector = loop_selector_component_type(paginator=paginator, parent=self)
        playhead_component_type = playhead_component_type or PlayheadComponent
        self._playhead = playhead_component_type(
            notes=playhead_notes,
            triplet_notes=playhead_triplet_notes,
            channels=playhead_channels,
            grid_resolution=(self._grid_resolution),
            paginator=paginator,
            parent=self,
        )

    @property
    def note_editor(self):
        return self._note_editor

    def set_pitch_provider(self, provider):
        self._note_editor.pitch_provider = provider

    def set_step_buttons(self, buttons):
        self._note_editor.set_matrix(buttons)

    def set_resolution_buttons(self, buttons):
        self._grid_resolution.resolution_buttons.set_control_element(buttons)

    def set_loop_buttons(self, matrix):
        self._loop_selector.set_matrix(matrix)

    def set_loop_delete_button(self, button):
        self._loop_selector.delete_button.set_control_element(button)

    def set_prev_page_button(self, button):
        self._loop_selector.prev_page_button.set_control_element(button)

    def set_next_page_button(self, button):
        self._loop_selector.next_page_button.set_control_element(button)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/view/notification.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4662 bytes
from math import inf
from typing import Any, Callable, Generic, Optional, Tuple
from uuid import uuid1

from ..event_signal import EventSignalFn, on_notification
from ..state import State
from ..type_decl import INIT_EVENT, ContentType, Event
from .type_decl import NotificationDataType, RenderNotification
from .util import suppress_notifications
from .view import View

DEFAULT_NOTIFICATION_DURATION: float = 2.0


class NotificationView(View[ContentType], Generic[(ContentType, NotificationDataType)]):
    def __init__(
        self,
        render_fn: RenderNotification[(Any, ContentType)],
        render_condition: Callable[([State], bool)] = lambda *_: True,
        notification_signal: EventSignalFn[NotificationDataType] = on_notification(),
        name: str = "notification",
        duration: Optional[float] = None,
        exclusive: bool = True,
        suppressing_signals: Tuple[(EventSignalFn[NotificationDataType], ...)] = (),
        supports_new_line: bool = False,
    ):
        self._render = render_fn
        self._render_condition = render_condition
        self._notification_signal = notification_signal
        self._name = name
        self._duration = duration
        self._exclusive = exclusive
        self._suppressing_signals = suppressing_signals
        self._supports_new_line = supports_new_line

    def init(self, state: State):
        if not hasattr(state, "_notifications"):
            state._notifications = set()
        if self._name in state._notifications:
            self._name += str(uuid1())
        state._notifications.add(self._name)
        self.reset_state(state)

    def reset_state(self, state: State, delay: Optional[float] = None):
        state.set_delayed(self._name, None, delay)

    def react(self, state: State, event: Event):
        if event is INIT_EVENT:
            self.init(state)
        notification_data = self._notification_signal(state, event)
        if notification_data:
            setattr(state, self._name, notification_data)
            if self._exclusive:
                suppress_notifications(state, exclude=[self._name])
            if self._duration != inf:
                self.reset_state(state, delay=(self._duration if self._duration else DEFAULT_NOTIFICATION_DURATION))
        else:
            if any(suppressing_signal(state, event) for suppressing_signal in self._suppressing_signals):
                self.reset_state(state)
        return state

    def render(self, state: State):
        notification_data = getattr(state, self._name)
        if not (self._supports_new_line or isinstance(notification_data, str)):
            return self._render(state, notification_data)
        return self._render(state, notification_data.replace("\n", " "))

    def render_condition(self, state: State) -> bool:
        notification_data = getattr(state, self._name)
        return notification_data is not None and self._render_condition(state, notification_data)

    def content_condition(self, content) -> bool:
        return True



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/Control.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 34364 bytes
from functools import partial

from ableton.v2.base import old_hasattr
from future.moves.itertools import zip_longest
from past.utils import old_div

from . import Task
from .Defaults import MOMENTARY_DELAY
from .SubjectSlot import SlotManager
from .Util import clamp, first, flatten, is_matrix, lazy_attribute, mixin, nop, product, second


class ControlManager(SlotManager):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._control_states = {}

    @lazy_attribute
    def _tasks(self):
        return Task.TaskGroup()

    def control_notifications_enabled(self):
        return True

    def update(self):
        for control_state in list(self._control_states.values()):
            control_state.update()


def control_event(event_name):
    def event_decorator(self):
        def event_listener_decorator(event_listener):
            self._event_listeners[event_name] = event_listener
            return self

        return event_listener_decorator

    return property(event_decorator)


class Control:
    value = control_event("value")

    class State(SlotManager):
        enabled = True

        def __init__(self, control=None, manager=None, channel=None, identifier=None, *a, **k):
            (super(Control.State, self).__init__)(*a, **k)
            self._manager = manager
            self._value_listener = control._event_listeners.get("value", None)
            self._event_listeners = control._event_listeners
            self._control_element = None
            self._value_slot = None
            self._channel = channel
            self._identifier = identifier
            self._register_value_slot(manager, control)
            manager.register_slot_manager(self)

        def set_control_element(self, control_element):
            self._control_element = control_element
            if self._control_element:
                self._control_element.reset_state()
                if self._channel is not None:
                    self._control_element.set_channel(self._channel)
                if self._identifier is not None:
                    self._control_element.set_identifier(self._identifier)
            if self._value_slot:
                self._value_slot.subject = control_element

        def _register_value_slot(self, manager, control):
            if self._event_listener_required():
                self._value_slot = self.register_slot(None, self._on_value, "value")

        def _event_listener_required(self):
            return len(self._event_listeners) > 0

        def _on_value(self, value, *a, **k):
            if self._value_listener:
                if self._notifications_enabled():
                    (self._value_listener)(self._manager, value, self, *a, **k)

        def _notifications_enabled(self):
            return self.enabled and self._manager.control_notifications_enabled()

        def update(self):
            pass

        def _get_channel(self):
            return self._channel

        def _set_channel(self, channel):
            self._channel = channel
            if self._control_element:
                self._control_element.set_channel(self._channel)

        channel = property(_get_channel, _set_channel)

        def _get_identifier(self):
            return self._identifier

        def _set_identifier(self, value):
            self._identifier = value
            if self._control_element:
                self._control_element.set_identifier(self._identifier)

        identifier = property(_get_identifier, _set_identifier)

    _extra_kws = {}
    _extra_args = []

    def __init__(self, extra_args=None, extra_kws=None, *a, **k):
        (super().__init__)(*a, **k)
        self._event_listeners = {}
        if extra_args is not None:
            self._extra_args = extra_args
        if extra_kws is not None:
            self._extra_kws = extra_kws

    def __get__(self, manager, owner):
        return self._get_state(manager)

    def __set__(self, manager, owner):
        raise RuntimeError("Cannot change control.")

    def _get_state(self, manager):
        if self not in manager._control_states:
            manager._control_states[self] = None
            manager._control_states[self] = (self.State)(self, manager, *(self._extra_args), **self._extra_kws)
        if manager._control_states[self] is None:
            raise RuntimeError("Cannot fetch state during construction of controls.")
        return manager._control_states[self]

    def _clear_state(self, manager):
        if self in manager._control_states:
            del manager._control_states[self]


class MappedControl(Control):
    class State(Control.State):
        def __init__(self, control=None, manager=None, *a, **k):
            (super(MappedControl.State, self).__init__)(control, manager, *a, **k)
            self._direct_mapping = None

        def set_control_element(self, control_element):
            if self._control_element:
                self._control_element.release_parameter()
            super(MappedControl.State, self).set_control_element(control_element)
            self._update_direct_connection()

        def _get_direct_mapping(self):
            return self._direct_mapping

        def _set_direct_mapping(self, direct_mapping):
            self._direct_mapping = direct_mapping
            self._update_direct_connection()

        mapped_parameter = property(_get_direct_mapping, _set_direct_mapping)

        def _update_direct_connection(self):
            if self._control_element:
                self._control_element.connect_to(self._direct_mapping)

        def _notifications_enabled(self):
            return super(MappedControl.State, self)._notifications_enabled() and self._direct_mapping is None

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


class ButtonControl(Control):
    DELAY_TIME = MOMENTARY_DELAY
    REPEAT_RATE = 0.1
    pressed = control_event("pressed")
    released = control_event("released")
    pressed_delayed = control_event("pressed_delayed")
    released_delayed = control_event("released_delayed")
    released_immediately = control_event("released_immediately")

    class State(Control.State):
        def __init__(
            self,
            control=None,
            manager=None,
            color=None,
            pressed_color=None,
            disabled_color=None,
            repeat=False,
            enabled=True,
            *a,
            **k,
        ):
            (super(ButtonControl.State, self).__init__)(control, manager, *a, **k)
            self._pressed_listener = control._event_listeners.get("pressed", None)
            self._released_listener = control._event_listeners.get("released", None)
            self._pressed_delayed_listener = control._event_listeners.get("pressed_delayed", None)
            self._released_delayed_listener = control._event_listeners.get("released_delayed", None)
            self._released_immediately_listener = control._event_listeners.get("released_immediately", None)
            self._repeat = repeat
            self._is_pressed = False
            self._enabled = enabled
            self._color = color if color is not None else "DefaultButton.On"
            self._disabled_color = disabled_color if disabled_color is not None else "DefaultButton.Disabled"
            self._pressed_color = pressed_color

        def _get_color(self):
            return self._color

        def _set_color(self, value):
            self._color = value
            self._send_current_color()

        color = property(_get_color, _set_color)

        def _get_pressed_color(self):
            return self._pressed_color

        def _set_pressed_color(self, value):
            self._pressed_color = value
            self._send_current_color()

        pressed_color = property(_get_pressed_color, _set_pressed_color)

        def _get_disabled_color(self):
            return self._disabled_color

        def _set_disabled_color(self, value):
            self._disabled_color = value
            self._send_current_color()

        disabled_color = property(_get_disabled_color, _set_disabled_color)

        def _get_enabled(self):
            return self._enabled

        def _set_enabled(self, enabled):
            if self._enabled != enabled:
                if not enabled:
                    self._release_button()
                self._enabled = enabled
                self._send_current_color()

        enabled = property(_get_enabled, _set_enabled)

        @property
        def is_momentary(self):
            return self._control_element and self._control_element.is_momentary()

        @property
        def is_pressed(self):
            return self._is_pressed

        def _event_listener_required(self):
            return True

        def set_control_element(self, control_element):
            if self._control_element != control_element:
                self._release_button()
            super(ButtonControl.State, self).set_control_element(control_element)
            self._send_current_color()

        def _send_current_color(self):
            if self._control_element:
                if not self._enabled:
                    self._control_element.set_light(self._disabled_color)
                else:
                    if self._pressed_color and self.is_pressed:
                        self._control_element.set_light(self._pressed_color)
                    else:
                        self._control_element.set_light(self._color)

        def _on_value(self, value, *a, **k):
            if self._notifications_enabled():
                if not self.is_momentary:
                    self._press_button()
                    self._release_button()
                else:
                    if value:
                        self._press_button()
                    else:
                        self._release_button()
                (super(ButtonControl.State, self)._on_value)(value, *a, **k)
            self._send_current_color()

        def _press_button(self):
            is_pressed = self._is_pressed
            self._is_pressed = True
            if self._notifications_enabled():
                if not is_pressed:
                    if self._pressed_listener is not None:
                        if self._repeat:
                            self._repeat_task.restart()
                        self._pressed_listener(self._manager, self)
                    if self._has_delayed_event():
                        self._delay_task.restart()

        def _release_button(self):
            is_pressed = self._is_pressed
            self._is_pressed = False
            if self._notifications_enabled():
                if is_pressed:
                    if self._released_listener is not None:
                        self._released_listener(self._manager, self)
                    if self._repeat:
                        self._repeat_task.kill()
                    if self._has_delayed_event():
                        if self._delay_task.is_running:
                            self._released_immediately_listener(self._manager, self)
                            self._delay_task.kill()
                        else:
                            self._released_delayed_listener(self._manager, self)

        @lazy_attribute
        def _delay_task(self):
            return self._manager._tasks.add(
                Task.sequence(Task.wait(ButtonControl.DELAY_TIME), Task.run(self._on_pressed_delayed)),
            )

        @lazy_attribute
        def _repeat_task(self):
            notify_pressed = partial(self._pressed_listener, self._manager, self)
            return self._manager._tasks.add(
                Task.sequence(
                    Task.wait(ButtonControl.DELAY_TIME),
                    Task.loop(Task.wait(ButtonControl.REPEAT_RATE), Task.run(notify_pressed)),
                ),
            )

        def _has_delayed_event(self):
            return (
                self._pressed_delayed_listener is not None
                or self._released_delayed_listener is not None
                or self._released_immediately_listener is not None
            )

        def _on_pressed_delayed(self):
            if self._notifications_enabled():
                if self._is_pressed:
                    self._pressed_delayed_listener(self._manager, self)

        def update(self):
            self._send_current_color()

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


class ToggleButtonControl(Control):
    toggled = control_event("toggled")

    class State(Control.State):
        def __init__(self, control=None, manager=None, untoggled_color=None, toggled_color=None, *a, **k):
            (super(ToggleButtonControl.State, self).__init__)(control, manager, *a, **k)
            self._untoggled_color = untoggled_color or "DefaultButton.Off"
            self._toggled_color = toggled_color or "DefaultButton.On"
            self._is_toggled = False
            self._toggled_listener = control._event_listeners.get("toggled", None)

        def _get_is_toggled(self):
            return self._is_toggled

        def _set_is_toggled(self, toggled):
            if self._is_toggled != toggled:
                self._is_toggled = toggled
                if self._toggled_listener:
                    if self._notifications_enabled():
                        self._toggled_listener(self._manager, self._is_toggled, self)
                self._send_current_color()

        is_toggled = property(_get_is_toggled, _set_is_toggled)

        def _get_untoggled_color(self):
            return self._untoggled_color

        def _set_untoggled_color(self, value):
            self._untoggled_color = value
            self._send_current_color()

        untoggled_color = property(_get_untoggled_color, _set_untoggled_color)

        def _get_toggled_color(self):
            return self._toggled_color

        def _set_toggled_color(self, value):
            self._toggled_color = value
            self._send_current_color()

        toggled_color = property(_get_toggled_color, _set_toggled_color)

        def set_control_element(self, control_element):
            super(ToggleButtonControl.State, self).set_control_element(control_element)
            self._send_current_color()

        def _send_current_color(self):
            if self._control_element:
                self._control_element.set_light(self._toggled_color if self._is_toggled else self._untoggled_color)

        def _is_momentary(self):
            return self._control_element and self._control_element.is_momentary()

        def _on_value(self, value, *a, **k):
            if self._notifications_enabled():
                if not (value or self._is_momentary()):
                    self._is_toggled = not self._is_toggled
                    if self._toggled_listener:
                        self._toggled_listener(self._manager, self._is_toggled, self)
                    self._send_current_color()
                (super(ToggleButtonControl.State, self)._on_value)(value, *a, **k)

        def update(self):
            self._send_current_color()

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


class RadioButtonControl(Control):
    checked = control_event("checked")

    class State(Control.State):
        def __init__(self, control, manager=None, unchecked_color=None, checked_color=None, *a, **k):
            (super(RadioButtonControl.State, self).__init__)(control, manager, *a, **k)
            self._unchecked_color = unchecked_color or "DefaultButton.Off"
            self._checked_color = checked_color or "DefaultButton.On"
            self._checked = False
            self._checked_listener = control._event_listeners.get("checked", None)
            self._on_checked = nop

        def _get_is_checked(self):
            return self._checked

        def _set_is_checked(self, value):
            if self._checked != value:
                self._checked = value
                self._notify_checked()
                self._send_current_color()

        is_checked = property(_get_is_checked, _set_is_checked)

        def _get_unchecked_color(self):
            return self._unchecked_color

        def _set_unchecked_color(self, value):
            self._unchecked_color = value
            self._send_current_color()

        unchecked_color = property(_get_unchecked_color, _set_unchecked_color)

        def _get_checked_color(self):
            return self._checked_color

        def _set_checked_color(self, value):
            self._checked_color = value
            self._send_current_color()

        checked_color = property(_get_checked_color, _set_checked_color)

        def set_control_element(self, control_element):
            super(RadioButtonControl.State, self).set_control_element(control_element)
            self._send_current_color()

        def _send_current_color(self):
            if self._control_element:
                self._control_element.set_light(self._checked_color if self._checked else self._unchecked_color)

        def _is_momentary(self):
            return self._control_element and self._control_element.is_momentary()

        def _on_value(self, value, *a, **k):
            if self._notifications_enabled():
                checked = value or self._is_momentary()
                if checked:
                    if not self._checked:
                        self.is_checked = True
                    (super(RadioButtonControl.State, self)._on_value)(value, *a, **k)

        def _notify_checked(self):
            if self._checked:
                if self._checked_listener:
                    if self._notifications_enabled():
                        self._checked_listener(self._manager, self)
                self._on_checked()

        def update(self):
            self._send_current_color()

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


class EncoderControl(Control):
    TOUCH_TIME = 0.5
    touched = control_event("touched")
    released = control_event("released")

    class State(Control.State):
        def __init__(self, control=None, manager=None, *a, **k):
            (super(EncoderControl.State, self).__init__)(control, manager, *a, **k)
            self._is_touched = False
            self._touched_listener = control._event_listeners.get("touched", None)
            self._released_listener = control._event_listeners.get("released", None)
            self._touch_value_slot = None
            self._timer_based = False
            if self._touched_listener or self._released_listener:
                self._touch_value_slot = self.register_slot(None, self._on_touch_value, "touch_value")

        @lazy_attribute
        def _release_task(self):
            return self._manager._tasks.add(
                Task.sequence(Task.wait(EncoderControl.TOUCH_TIME), Task.run(self._release_encoder)),
            )

        @property
        def is_touched(self):
            return self._is_touched

        def set_control_element(self, control_element):
            if self._control_element != control_element:
                self._release_encoder()
            super(EncoderControl.State, self).set_control_element(control_element)
            if self._touch_value_slot:
                self._touch_value_slot.subject = control_element
            if control_element:
                if old_hasattr(control_element, "is_pressed"):
                    if control_element.is_pressed():
                        self._touch_encoder()

        def _touch_encoder(self):
            is_touched = self._is_touched
            self._is_touched = True
            if self._notifications_enabled():
                if self._touched_listener:
                    if not is_touched:
                        self._touched_listener(self._manager, self)

        def _release_encoder(self):
            is_touched = self._is_touched
            self._is_touched = False
            self._timer_based = False
            if self._notifications_enabled():
                if self._released_listener:
                    if is_touched:
                        self._released_listener(self._manager, self)

        def _on_value(self, value, *a, **k):
            if self._notifications_enabled():
                if not self._is_touched:
                    self._timer_based = True
                    self._touch_encoder()
                    self._release_task.restart()
                else:
                    if self._timer_based:
                        self._release_task.restart()
                if self._value_listener:
                    if self._control_element:
                        normalized_value = self._control_element.normalize_value(value)
                        (self._value_listener)(self._manager, normalized_value, self, *a, **k)

        def _on_touch_value(self, value, *a, **k):
            if self._notifications_enabled():
                if value:
                    self._touch_encoder()
                else:
                    self._release_encoder()


class PlayableControl(ButtonControl):
    class State(ButtonControl.State):
        def __init__(self, *a, **k):
            (super(PlayableControl.State, self).__init__)(*a, **k)
            self._enabled = True
            self._playable = True

        def set_control_element(self, control_element):
            super(PlayableControl.State, self).set_control_element(control_element)
            self._update_script_forwarding()
            self._send_current_color()

        def _update_script_forwarding(self):
            if self._control_element:
                if self._enabled:
                    self._control_element.suppress_script_forwarding = self._playable

        def _get_enabled(self):
            return self._enabled

        def _set_enabled(self, enabled):
            super(PlayableControl.State, self)._set_enabled(enabled)
            if not enabled or self._control_element:
                self._control_element.reset_state()
                self._send_current_color()
            else:
                self.set_control_element(self._control_element)

        enabled = property(_get_enabled, _set_enabled)

        def set_playable(self, value):
            self._playable = value
            self._update_script_forwarding()

        def _notifications_enabled(self):
            return super(PlayableControl.State, self)._notifications_enabled() and not self._playable


_DYNAMIC_CONTROL_COUNT = None


class ControlList(Control):
    DYNAMIC_CONTROL_COUNT = _DYNAMIC_CONTROL_COUNT

    class State(Control.State):
        _extra_kws = {}
        _extra_args = []

        def __init__(
            self,
            control=None,
            manager=None,
            extra_args=None,
            extra_kws=None,
            unavailable_color=None,
            *a,
            **k,
        ):
            (super(ControlList.State, self).__init__)(a, manager=manager, control=control, **k)
            self._control_elements = None
            self._control_type = control.control_type
            self._controls = []
            self._dynamic_create = False
            self._unavailable_color = unavailable_color if unavailable_color is not None else "DefaultButton.Disabled"
            if extra_args is not None:
                self._extra_args = extra_args
            if extra_kws is not None:
                self._extra_kws = extra_kws
            self.control_count = control.control_count

        def _get_control_count(self):
            return len(self._controls)

        def _set_control_count(self, count):
            self._dynamic_create = count == ControlList.DYNAMIC_CONTROL_COUNT
            if self._dynamic_create:
                count = len(self._control_elements) if self._control_elements else 0
            self._create_controls(count)
            self._update_controls()

        control_count = property(_get_control_count, _set_control_count)

        def _get_unavailable_color(self):
            return self._unavailable_color

        def _set_unavailable_color(self, value):
            self._unavailable_color = value
            control_elements = self._control_elements or []
            for control, element in zip_longest(self._controls, control_elements):
                if not control:
                    if element:
                        self._send_unavailable_color(element)

        unavailable_color = property(_get_unavailable_color, _set_unavailable_color)

        def _create_controls(self, count):
            self._disconnect_controls()
            self._controls = [self._make_control(i) for i in range(count)]

        def _disconnect_controls(self):
            for control in self._controls:
                control._get_state(self._manager).disconnect()
                control._clear_state(self._manager)

        def _make_control(self, index):
            control = (self._control_type)(*self._extra_args, **self._extra_kws)
            control._event_listeners = self._event_listeners
            control_state = control._get_state(self._manager)
            if not old_hasattr(control_state, "index"):
                control_state.index = index
            else:
                raise RuntimeError("Cannot set 'index' attribute. Attribute already set.")
            return control

        def set_control_element(self, control_elements):
            self._control_elements = control_elements
            if self._dynamic_create:
                if len(control_elements or []) != len(self._control_element or []):
                    self._create_controls(len(control_elements or []))
            self._update_controls()

        def _update_controls(self):
            control_elements = self._control_elements or []
            for control, element in zip_longest(self._controls, control_elements):
                if control:
                    control._get_state(self._manager).set_control_element(element)
                if element:
                    element.reset_state()
                    self._send_unavailable_color(element)

        def _send_unavailable_color(self, element):
            if old_hasattr(element, "set_light"):
                element.set_light(self._unavailable_color)

        def __getitem__(self, index):
            return self._controls[index]._get_state(self._manager)

        def _on_value(self, value, *a, **k):
            pass

        def _register_value_slot(self, manager, control):
            pass

    def __init__(self, control_type=None, control_count=_DYNAMIC_CONTROL_COUNT, *a, **k):
        (super().__init__)(a, extra_args=a, extra_kws=k, **k)
        self.control_type = control_type
        self.control_count = control_count


class RadioButtonGroup(ControlList, RadioButtonControl):
    class State(ControlList.State):
        def __init__(self, *a, **k):
            self._checked_index = 0
            (super(RadioButtonGroup.State, self).__init__)(*a, **k)

        def _create_controls(self, count):
            self._checked_index = clamp(self._checked_index, 0, count - 1)
            super(RadioButtonGroup.State, self)._create_controls(count)

        def _make_control(self, index):
            control = super(RadioButtonGroup.State, self)._make_control(index)
            control_state = control._get_state(self._manager)
            control_state._on_checked = partial(self._on_checked, control_state)
            control_state.is_checked = index == self._checked_index
            return control

        def _on_checked(self, checked_control):
            for control in self._controls:
                control = control._get_state(self._manager)
                control.is_checked = control == checked_control

            self._checked_index = checked_control.index

    def __init__(self, *a, **k):
        (super().__init__)(RadioButtonControl, *a, **k)


_DYNAMIC_MATRIX_DIMENSIONS = (None, None)


class MatrixControl(ControlList):
    DYNAMIC_DIMENSIONS = _DYNAMIC_MATRIX_DIMENSIONS

    class State(ControlList.State):
        def __init__(self, control=None, manager=None, *a, **k):
            self._dimensions = (None, None)
            (super(MatrixControl.State, self).__init__)(control, manager, *a, **k)

        def _get_dimensions(self):
            return self._dimensions

        def _set_dimensions(self, dimensions):
            self._dynamic_create = dimensions == MatrixControl.DYNAMIC_DIMENSIONS
            if self._dynamic_create:
                count = len(self._control_elements) if self._control_elements else 0
            self._dimensions = dimensions
            count = first(dimensions) * second(dimensions)
            self._create_controls(count)
            self._update_controls()

        dimensions = property(_get_dimensions, _set_dimensions)

        def _make_control(self, index):
            control = super(MatrixControl.State, self)._make_control(index)
            control_state = control._get_state(self._manager)
            if not old_hasattr(control_state, "coordinate"):
                control_state.coordinate = (int(old_div(index, self.width)), index % self.width)
            else:
                raise RuntimeError("Cannot set 'coordinate' attribute. Attribute already set.")
            return control

        def set_control_element(self, control_elements):
            dimensions = (None, None)
            if old_hasattr(control_elements, "width") and old_hasattr(control_elements, "height"):
                dimensions = (control_elements.height(), control_elements.width())
                if not self._dynamic_create:
                    control_elements = [
                        control_elements.get_button(col, row)
                        for row, col in product(range(self.height), range(self.width))
                    ]
            else:
                if is_matrix(control_elements):
                    dimensions = (len(control_elements), len(first(control_elements)))
                    if not self._dynamic_create:
                        control_elements = [row[0 : self.width] for row in control_elements]
                    control_elements = list(flatten(control_elements))
                else:
                    if control_elements is not None:
                        raise RuntimeError("Control Elements must be a matrix")
            if self._dynamic_create:
                if None not in dimensions:
                    self._dimensions = dimensions
                    self._create_controls(first(dimensions) * second(dimensions))
                    self._update_controls()
            super(MatrixControl.State, self).set_control_element(control_elements)

        def get_control(self, row, column):
            index = row * self.width + column
            return self._controls[index]._get_state(self._manager)

        @property
        def width(self):
            return second(self._dimensions)

        @property
        def height(self):
            return first(self._dimensions)

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)


_control_list_classes = {}
_control_matrix_classes = {}


def control_list(control_type, *a, **k):
    if control_type == RadioButtonControl:
        return RadioButtonGroup(*a, **k)
    c = _control_list_classes.get(control_type)
    if not c:
        c = mixin(ControlList, control_type)
        c.State = mixin(ControlList.State, control_type.State)
        _control_list_classes[control_type] = c
    return c(control_type, *a, **k)


def control_matrix(control_type, *a, **k):
    m = _control_matrix_classes.get(control_type)
    if not m:
        m = mixin(MatrixControl, control_type)
        m.State = mixin(MatrixControl.State, control_type.State)
        _control_matrix_classes[control_type] = m
    return m(control_type, *a, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/drum_group.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 11924 bytes
from ...base import clamp, depends, find_if, first, listens, listens_group, liveobj_changed, liveobj_valid
from ..control import ButtonControl
from .playable import PlayableComponent
from .slide import Slideable, SlideComponent

BASE_DRUM_RACK_NOTE = 36


class DrumGroupComponent(PlayableComponent, SlideComponent, Slideable):
    mute_button = ButtonControl()
    solo_button = ButtonControl()
    delete_button = ButtonControl()
    quantize_button = ButtonControl()

    @depends(set_pad_translations=None)
    def __init__(self, translation_channel=None, set_pad_translations=None, *a, **k):
        self._drum_group_device = None
        self._selected_drum_pad = None
        self._all_drum_pads = []
        self._assigned_drum_pads = []
        self._translation_channel = translation_channel
        (super().__init__)(*a, **k)
        self._set_pad_translations = set_pad_translations

    position_count = 32
    page_length = 4
    page_offset = 1

    def contents_range(self, pmin, pmax):
        pos_count = self.position_count
        first_pos = max(int(pmin - 0.05), 0)
        last_pos = min(int(pmax + 0.2), pos_count)
        return list(range(first_pos, last_pos))

    def contents(self, index):
        drum = self._drum_group_device
        if liveobj_valid(drum):
            return any(pad.chains for pad in drum.drum_pads[index * 4 : index * 4 + 4])
        return False

    @property
    def position(self):
        if liveobj_valid(self._drum_group_device):
            return self._drum_group_device.view.drum_pads_scroll_position
        return 0

    @position.setter
    def position(self, index):
        if liveobj_valid(self._drum_group_device):
            self._drum_group_device.view.drum_pads_scroll_position = index

    @property
    def assigned_drum_pads(self):
        return self._assigned_drum_pads

    @property
    def min_pitch(self):
        if self.assigned_drum_pads:
            return self.assigned_drum_pads[0].note
        return BASE_DRUM_RACK_NOTE

    @property
    def max_pitch(self):
        if self.assigned_drum_pads:
            return self.assigned_drum_pads[-1].note
        return BASE_DRUM_RACK_NOTE

    def _update_assigned_drum_pads(self):
        assigned_drum_pads = []
        visible_drum_pads = (
            self._drum_group_device.visible_drum_pads if liveobj_valid(self._drum_group_device) else None
        )
        if visible_drum_pads:
            if self._all_drum_pads:
                first_pad = first(visible_drum_pads)
                if first_pad:
                    size = self.width * self.height
                    first_note = first_pad.note
                    if first_note > 128 - size:
                        size = 128 - first_note
                    offset = clamp(first_note, 0, 128 - len(visible_drum_pads))
                    assigned_drum_pads = self._all_drum_pads[offset : offset + size]
        self._assigned_drum_pads = assigned_drum_pads

    def set_matrix(self, matrix):
        super().set_matrix(matrix)
        self._update_assigned_drum_pads()
        self._create_and_set_pad_translations()

    def set_drum_group_device(self, drum_group_device):
        if drum_group_device:
            if not drum_group_device.can_have_drum_pads:
                drum_group_device = None
            if drum_group_device != self._drum_group_device:
                self._DrumGroupComponent__on_visible_drum_pads_changed.subject = drum_group_device
                drum_group_view = drum_group_device.view if drum_group_device else None
                self._DrumGroupComponent__on_selected_drum_pad_changed.subject = drum_group_view
                self._DrumGroupComponent__on_drum_pads_scroll_position_changed.subject = drum_group_view
                self._DrumGroupComponent__on_chains_changed.subject = drum_group_device
                self._drum_group_device = drum_group_device
                self._update_drum_pad_listeners()
                self._update_selected_drum_pad()
                self._update_note_translations()
                super().update()

    def update(self):
        super().update()
        if self.is_enabled():
            self.notify_position()

    def _update_drum_pad_listeners(self):
        if liveobj_valid(self._drum_group_device):
            self._all_drum_pads = self._drum_group_device.drum_pads
            visible_drum_pads = self._drum_group_device.visible_drum_pads
            self._DrumGroupComponent__on_solo_changed.replace_subjects(visible_drum_pads)
            self._DrumGroupComponent__on_mute_changed.replace_subjects(visible_drum_pads)
            self._update_assigned_drum_pads()
            self._update_note_translations()

    @listens_group("solo")
    def __on_solo_changed(self, pad):
        self._update_led_feedback()

    @listens_group("mute")
    def __on_mute_changed(self, pad):
        self._update_led_feedback()

    def _update_led_feedback(self):
        if liveobj_valid(self._drum_group_device):
            super()._update_led_feedback()

    def _update_button_color(self, button):
        pad = self._pad_for_button(button)
        button.color = self._color_for_pad(pad) if pad else "DefaultButton.On"

    def _color_for_pad(self, pad):
        has_soloed_pads = bool(find_if(lambda pad: pad.solo, self._all_drum_pads))
        button_color = "DrumGroup.PadEmpty"
        if pad == self._selected_drum_pad:
            button_color = "DrumGroup.PadSelected"
            if has_soloed_pads and not pad.solo or pad.mute:
                button_color = "DrumGroup.PadSelectedNotSoloed"
            else:
                if pad.mute and not pad.solo:
                    button_color = "DrumGroup.PadMutedSelected"
                else:
                    if not has_soloed_pads or pad.solo:
                        button_color = "DrumGroup.PadSoloedSelected"
        else:
            if pad.chains:
                button_color = "DrumGroup.PadFilled"
                if has_soloed_pads and not pad.solo:
                    button_color = "DrumGroup.PadFilled" if (not pad.mute) else "DrumGroup.PadMuted"
                else:
                    if not has_soloed_pads or pad.mute:
                        button_color = "DrumGroup.PadMuted"
                    else:
                        if has_soloed_pads:
                            if pad.solo:
                                button_color = "DrumGroup.PadSoloed"
        return button_color

    def _button_coordinates_to_pad_index(self, first_note, coordinates):
        y, x = coordinates
        y = self.height - y - 1
        if x < 4 and y >= 4:
            first_note += 16
        else:
            if x >= 4 and y < 4:
                first_note += 4 * self.width
            else:
                if x >= 4:
                    if y >= 4:
                        first_note += 4 * self.width + 16
        return x % 4 + y % 4 * 4 + first_note

    def _on_matrix_pressed(self, button):
        selected_drum_pad = self._pad_for_button(button)
        if self.mute_button.is_pressed:
            selected_drum_pad.mute = not selected_drum_pad.mute
        if self.solo_button.is_pressed:
            selected_drum_pad.solo = not selected_drum_pad.solo
        if self.quantize_button.is_pressed:
            button.color = "DrumGroup.PadAction"
            self.quantize_pitch(selected_drum_pad.note)
        if self.delete_button.is_pressed:
            button.color = "DrumGroup.PadAction"
            self.delete_pitch(selected_drum_pad)
        if self.select_button.is_pressed:
            self._drum_group_device.view.selected_drum_pad = selected_drum_pad
            self.select_drum_pad(selected_drum_pad)
            super()._on_matrix_pressed(button)
        if self.mute_button.is_pressed or self.solo_button.is_pressed:
            self._update_led_feedback()

    @listens("visible_drum_pads")
    def __on_visible_drum_pads_changed(self):
        self._update_drum_pad_listeners()
        self._update_led_feedback()

    @listens("drum_pads_scroll_position")
    def __on_drum_pads_scroll_position_changed(self):
        self._update_note_translations()
        self._update_led_feedback()
        self.notify_position()

    @listens("selected_drum_pad")
    def __on_selected_drum_pad_changed(self):
        self._update_selected_drum_pad()

    @listens("chains")
    def __on_chains_changed(self):
        self._update_led_feedback()

    def _update_selected_drum_pad(self):
        selected_drum_pad = (
            self._drum_group_device.view.selected_drum_pad if liveobj_valid(self._drum_group_device) else None
        )
        if liveobj_changed(self._selected_drum_pad, selected_drum_pad):
            self._selected_drum_pad = selected_drum_pad
            self._update_led_feedback()
            self._on_selected_drum_pad_changed()

    def _on_selected_drum_pad_changed(self):
        pass

    @mute_button.value
    def mute_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @solo_button.value
    def solo_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @delete_button.value
    def delete_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @quantize_button.value
    def quantize_button(self, value, button):
        self._set_control_pads_from_script(bool(value))

    @property
    def has_assigned_pads(self):
        return self._assigned_drum_pads and liveobj_valid(first(self._assigned_drum_pads))

    def _pad_for_button(self, button):
        if self.has_assigned_pads:
            index = self._button_coordinates_to_pad_index(first(self._assigned_drum_pads).note, button.coordinate)
            if index < 128:
                return self._all_drum_pads[index]
            return None
        return None

    def _note_translation_for_button(self, button):
        identifier = None
        channel = None
        if self.has_assigned_pads:
            identifier = self._button_coordinates_to_pad_index(first(self._assigned_drum_pads).note, button.coordinate)
            channel = self._translation_channel
        return (identifier, channel)

    def _update_note_translations(self):
        if self._assigned_drum_pads:
            if not self._can_set_pad_translations():
                super()._update_note_translations()

    def _can_set_pad_translations(self):
        return self.width <= 4 and self.height <= 4

    def _create_and_set_pad_translations(self):
        def create_translation_entry(button):
            row, col = button.coordinate
            return (col, row, button.identifier, button.channel)

        if self._can_set_pad_translations():
            translations = []
            for button in self.matrix:
                button.channel = self._translation_channel
                button.identifier = self._button_coordinates_to_pad_index(BASE_DRUM_RACK_NOTE, button.coordinate)
                button.enabled = True
                translations.append(create_translation_entry(button))

            self._set_pad_translations(tuple(translations))
        else:
            self._update_note_translations()
            self._set_pad_translations(None)

    def select_drum_pad(self, drum_pad):
        raise NotImplementedError

    def quantize_pitch(self, note):
        raise NotImplementedError

    def delete_pitch(self, drum_pad):
        raise NotImplementedError



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/mode/mode.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3143 bytes
from ...base import depends, nop, task
from . import LayerModeBase, Mode
from .modes import ModesComponent


class EnablingAddLayerMode(LayerModeBase):
    @depends(parent_task_group=None)
    def __init__(self, parent_task_group=None, *a, **k):
        (super().__init__)(*a, **k)
        self._should_track_layers = (
            not isinstance(self._component, ModesComponent)
            and not hasattr(self._component, "disable_layer_tracking")
            and hasattr(self._component, "num_layers")
        )
        if self._should_track_layers:
            self._possibly_disable_component_task = parent_task_group.add(task.run(self._possibly_disable_component))
            self._possibly_disable_component_task.kill()

    def enter_mode(self):
        component = self._get_component()
        self._layer.grab(component)
        if not component.is_enabled():
            component.set_enabled(True)

    def leave_mode(self):
        component = self._get_component()
        self._layer.release(component)
        if self._should_track_layers:
            self._possibly_disable_component_task.restart()

    def _possibly_disable_component(self):
        component = self._get_component()
        if not component.num_layers:
            component.set_enabled(False)


class CallFunctionMode(Mode):
    def __init__(self, on_enter_fn=nop, on_exit_fn=nop, *a, **k):
        (super().__init__)(*a, **k)
        self._on_enter_fn = on_enter_fn
        self._on_exit_fn = on_exit_fn

    def enter_mode(self):
        self._on_enter_fn()

    def leave_mode(self):
        self._on_exit_fn()


class ShowDetailClipMode(Mode):
    def __init__(self, application, *a, **k):
        (super().__init__)(*a, **k)
        self._view = application.view
        self._needs_restore_view = False

    def enter_mode(self):
        self._needs_restore_view = self._view.is_view_visible("Detail/DeviceChain", False)
        if self._needs_restore_view:
            self._view.show_view("Detail/Clip")

    def leave_mode(self):
        if self._needs_restore_view:
            self._view.show_view("Detail/DeviceChain")
        self._needs_restore_view = False



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ViewControlComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5974 bytes
import Live

from .CompoundComponent import CompoundComponent
from .Dependency import depends
from .ScrollComponent import Scrollable, ScrollComponent
from .Util import in_range

NavDirection = Live.Application.Application.View.NavDirection
VIEWS = ("Browser", "Arranger", "Session", "Detail", "Detail/Clip", "Detail/DeviceChain")


class _DeltaSongScroller(Scrollable):
    @depends(song=None)
    def __init__(self, song=None, *a, **k):
        (super().__init__)(*a, **k)
        self._song = song

    _do_scroll = NotImplemented
    _can_scroll = NotImplemented

    def scroll_up(self):
        if self.can_scroll_up():
            self._do_scroll(-1)

    def scroll_down(self):
        if self.can_scroll_down():
            self._do_scroll(1)

    def can_scroll_up(self):
        return self._can_scroll(-1)

    def can_scroll_down(self):
        return self._can_scroll(1)


def tracks_to_use(song):
    return list(tuple(song.visible_tracks) + tuple(song.return_tracks) + (song.master_track,))


def next_item(seq, item, delta):
    return seq[list(seq).index(item) + delta]


def has_next_item(seq, item, delta):
    try:
        return in_range(list(seq).index(item) + delta, 0, len(seq))
    except ValueError:
        return False


class TrackScroller(_DeltaSongScroller):
    def _do_scroll(self, delta):
        song = self._song
        tracks = tracks_to_use(song)
        track = next_item(tracks, song.view.selected_track, delta)
        song.view.selected_track = track
        if track.can_be_armed:
            playing_slot_index = track.playing_slot_index
            if playing_slot_index >= 0:
                if track.clip_slots[playing_slot_index].clip:
                    song.view.highlighted_clip_slot = track.clip_slots[playing_slot_index]

    def _can_scroll(self, delta):
        tracks = tracks_to_use(self._song)
        try:
            return has_next_item(tracks, self._song.view.selected_track, delta)
        except ValueError:
            return False


class BasicSceneScroller(_DeltaSongScroller):
    def _do_scroll(self, delta):
        song = self._song
        view = song.view
        view.selected_scene = next_item(song.scenes, view.selected_scene, delta)

    def _can_scroll(self, delta):
        song = self._song
        view = song.view
        return has_next_item(song.scenes, view.selected_scene, delta)


class SceneScroller(BasicSceneScroller):
    def _do_scroll(self, delta):
        super()._do_scroll(delta)
        if self._song.view.highlighted_clip_slot is not None:
            if self._song.view.highlighted_clip_slot.has_clip:
                self._song.view.highlighted_clip_slot.fire(
                    force_legato=True,
                    launch_quantization=(Live.Song.Quantization.q_no_q),
                )
            else:
                self._song.view.selected_track.stop_all_clips(False)


class SceneListScroller(BasicSceneScroller):
    def _do_scroll(self, delta):
        super()._do_scroll(delta)
        self._song.view.selected_scene.fire(force_legato=True, can_select_scene_on_launch=False)


class ViewControlComponent(CompoundComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._scroll_tracks, self._scroll_scene_list, self._scroll_scenes = self.register_components(
            ScrollComponent(TrackScroller()),
            ScrollComponent(SceneListScroller()),
            ScrollComponent(SceneScroller()),
        )
        song = self.song()
        view = song.view
        self.register_slot(song, self._scroll_tracks.update, "visible_tracks")
        self.register_slot(song, self._scroll_tracks.update, "return_tracks")
        self.register_slot(song, self._scroll_scenes.update, "scenes")
        self.register_slot(song, self._scroll_scene_list.update, "scenes")
        self.register_slot(view, self._scroll_tracks.update, "selected_track")
        self.register_slot(view, self._scroll_scenes.update, "selected_scene")
        self.register_slot(view, self._scroll_scene_list.update, "selected_scene")

    def set_next_track_button(self, button):
        self._scroll_tracks.set_scroll_down_button(button)

    def set_prev_track_button(self, button):
        self._scroll_tracks.set_scroll_up_button(button)

    def set_next_scene_button(self, button):
        self._scroll_scenes.set_scroll_down_button(button)

    def set_prev_scene_button(self, button):
        self._scroll_scenes.set_scroll_up_button(button)

    def set_next_scene_list_button(self, button):
        self._scroll_scene_list.set_scroll_down_button(button)

    def set_prev_scene_list_button(self, button):
        self._scroll_scene_list.set_scroll_up_button(button)

    def show_view(self, view):
        app_view = self.application().view
        try:
            if view != "Detail/DeviceChain":
                pass
            if not app_view.is_view_visible("Detail"):
                app_view.show_view("Detail")
            if not app_view.is_view_visible(view):
                app_view.show_view(view)
        except RuntimeError:
            pass

    def focus_view(self, view):
        app_view = self.application().view
        if view != "Detail/DeviceChain":
            pass
        if not app_view.is_view_visible("Detail"):
            app_view.show_view("Detail")
        if not app_view.is_view_visible(view):
            app_view.focus_view(view)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/discrete_values.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2385 bytes
from ...base import nop
from .. import MIDI_CC_TYPE, InputControlElement, NotifyingControlElement


class ValueElement(NotifyingControlElement):
    def __init__(self, *a, **k):
        (super().__init__)(a, is_private=True, **k)
        self.send_value = nop
        self.set_light = nop
        self.is_momentary = lambda: False

    def reset(self):
        pass


class DiscreteValuesElement(InputControlElement):
    def __init__(self, identifier, channel=0, msg_type=MIDI_CC_TYPE, values=None, *a, **k):
        (super().__init__)(a, **k)
        self._sub_elements = {v: ValueElement() for v in values}

    def script_wants_forwarding(self):
        return True

    def message_map_mode(self):
        raise AssertionError("DiscreteValuesElement doesn't support mapping.")

    def receive_value(self, value):
        super().receive_value(value)
        if value in self._sub_elements:
            self._sub_elements[value].notify_value(127)

    def __iter__(self):
        yield from self._sub_elements.values()

    def __getitem__(self, value):
        return self._sub_elements[value]

    def __len__(self):
        return len(self._sub_elements)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements/touch.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1886 bytes
from ...base import listenable_property, listens
from ...live import song
from . import ButtonElement


class TouchElement(ButtonElement):
    __events__ = ("assignment",)

    def __init__(self, *a, encoder=None, **k):
        (super().__init__)(*a, **k)
        self._TouchElement__on_parameter_changed.subject = encoder
        self._TouchElement__on_parameter_name_changed.subject = encoder
        self._TouchElement__on_parameter_value_changed.subject = encoder
        self._encoder = encoder

    @listenable_property
    def controlled_parameter(self):
        return self._encoder.mapped_object

    def receive_value(self, value):
        super().receive_value(value)
        if value:
            song().begin_undo_step()
        else:
            song().end_undo_step()

    @listens("parameter")
    def __on_parameter_changed(self):
        self.notify_assignment()

    @listens("parameter_name")
    def __on_parameter_name_changed(self):
        self.notify_controlled_parameter()

    @listens("parameter_value")
    def __on_parameter_value_changed(self):
        self.notify_controlled_parameter()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/type_decl.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1109 bytes
from typing import Any, Callable, NamedTuple, TypeVar

from .state import State


class Event(NamedTuple):
    name: str
    origin: Any
    value: Any


INIT_EVENT = Event(name="init", origin=None, value=None)
DISCONNECT_EVENT = Event(name="disconnect", origin=None, value=None)
ContentType = TypeVar("ContentType")
Render = Callable[([State], ContentType)]



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class Application:
    """
    This class represents the Live application.
    """

    class View:
        """
        This class represents the view aspects of the Live application.
        """

        class NavDirection:
            """
            Class that represent an enumeration of values for NavDirection
            int([x]) -> integer
            int(x, base=10) -> integer

            Convert a number or string to an integer, or return 0 if no arguments
            are given.  If x is a number, return x.__int__().  For floating point
            numbers, this truncates towards zero.

            If x is not a number or if base is given, then x must be a string,
            bytes, or bytearray instance representing an integer literal in the
            given base.  The literal can be preceded by '+' or '-' and be surrounded
            by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
            Base 0 means to interpret the base from the string as an integer literal.
            >>> int('0b100', base=0)
            4
            """

            up = 0
            down = 1
            left = 2
            right = 3

        def add_browse_mode_listener(self, listener: Callable) -> None:
            """
            add_browse_mode_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "browse_mode" has changed.

                C++ signature :
                    void add_browse_mode_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def add_focused_document_view_listener(self, listener: Callable) -> None:
            """
            add_focused_document_view_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "focused_document_view" has changed.

                C++ signature :
                    void add_focused_document_view_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def add_is_view_visible_listener(self, arg2: object, listener: Callable) -> None:
            """
            add_is_view_visible_listener( (View)arg1, (object)arg2, (object)arg3) -> None :
                Add a listener function or method, which will be called as soon as the
                property "is_view_visible" has changed.

                C++ signature :
                    void add_is_view_visible_listener(TPyViewData<ASongApp>,TString,boost::python::api::object)
            """

        def add_view_focus_changed_listener(self, listener: Callable) -> None:
            """
            add_view_focus_changed_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "view_focus_changed" has changed.

                C++ signature :
                    void add_view_focus_changed_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def available_main_views(self) -> StringVector:
            """
            available_main_views( (View)arg1) -> StringVector :
                Return a list of strings with the available subcomponent views, which
                is to be specified, when using the rest of this classes functions.
                A 'subcomponent view' is a main view component of a document view, like
                the Session view, the Arranger or Detailview and so on...

                C++ signature :
                    std::__1::vector<TString, std::__1::allocator<TString>> available_main_views(TPyViewData<ASongApp>)
            """

        def browse_mode_has_listener(self, listener: Callable) -> bool:
            """
            browse_mode_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "browse_mode".

                C++ signature :
                    bool browse_mode_has_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def focus_view(self, arg2: object) -> None:
            """
            focus_view( (View)arg1, (object)arg2) -> None :
                Show and focus one through the identifier string specified view.

                C++ signature :
                    void focus_view(TPyViewData<ASongApp>,TString)
            """

        def focused_document_view_has_listener(self, listener: Callable) -> bool:
            """
            focused_document_view_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "focused_document_view".

                C++ signature :
                    bool focused_document_view_has_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def hide_view(self, arg2: object) -> None:
            """
            hide_view( (View)arg1, (object)arg2) -> None :
                Hide one through the identifier string specified view.

                C++ signature :
                    void hide_view(TPyViewData<ASongApp>,TString)
            """

        def is_view_visible(self, identifier: object, main_window_only: bool = True) -> bool:
            """
            is_view_visible( (View)arg1, (object)identifier [, (bool)main_window_only=True]) -> bool :
                Return true if the through the identifier string specified view is currently
                visible. If main_window_only is set to False, this will also check in second
                window. Notifications from the second window are not yet supported.

                C++ signature :
                    bool is_view_visible(TPyViewData<ASongApp>,TString [,bool=True])
            """

        def is_view_visible_has_listener(self, arg2: object, listener: Callable) -> bool:
            """
            is_view_visible_has_listener( (View)arg1, (object)arg2, (object)arg3) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "is_view_visible".

                C++ signature :
                    bool is_view_visible_has_listener(TPyViewData<ASongApp>,TString,boost::python::api::object)
            """

        def remove_browse_mode_listener(self, listener: Callable) -> None:
            """
            remove_browse_mode_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "browse_mode".

                C++ signature :
                    void remove_browse_mode_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def remove_focused_document_view_listener(self, listener: Callable) -> None:
            """
            remove_focused_document_view_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "focused_document_view".

                C++ signature :
                    void remove_focused_document_view_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def remove_is_view_visible_listener(self, arg2: object, listener: Callable) -> None:
            """
            remove_is_view_visible_listener( (View)arg1, (object)arg2, (object)arg3) -> None :
                Remove a previously set listener function or method from
                property "is_view_visible".

                C++ signature :
                    void remove_is_view_visible_listener(TPyViewData<ASongApp>,TString,boost::python::api::object)
            """

        def remove_view_focus_changed_listener(self, listener: Callable) -> None:
            """
            remove_view_focus_changed_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "view_focus_changed".

                C++ signature :
                    void remove_view_focus_changed_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def scroll_view(self, arg2: int, arg3: object, arg4: bool) -> None:
            """
            scroll_view( (View)arg1, (int)arg2, (object)arg3, (bool)arg4) -> None :
                Scroll through the identifier string specified view into the given
                direction, if possible.  Will silently return if the specified view
                can not perform the requested action.

                C++ signature :
                    void scroll_view(TPyViewData<ASongApp>,int,TString,bool)
            """

        def show_view(self, arg2: object) -> None:
            """
            show_view( (View)arg1, (object)arg2) -> None :
                Show one through the identifier string specified view. Will throw a
                runtime error if this is called in Live's initialization scope.

                C++ signature :
                    void show_view(TPyViewData<ASongApp>,TString)
            """

        def toggle_browse(self) -> None:
            """
            toggle_browse( (View)arg1) -> None :
                Reveals the device chain, the browser and starts hot swap for
                the selected device. Calling this function again stops hot swap.

                C++ signature :
                    void toggle_browse(TPyViewData<ASongApp>)
            """

        def view_focus_changed_has_listener(self, listener: Callable) -> bool:
            """
            view_focus_changed_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "view_focus_changed".

                C++ signature :
                    bool view_focus_changed_has_listener(TPyViewData<ASongApp>,boost::python::api::object)
            """

        def zoom_view(self, arg2: int, arg3: object, arg4: bool) -> None:
            """
            zoom_view( (View)arg1, (int)arg2, (object)arg3, (bool)arg4) -> None :
                Zoom through the identifier string specified view into the given
                direction, if possible.  Will silently return if the specified view
                can not perform the requested action.

                C++ signature :
                    void zoom_view(TPyViewData<ASongApp>,int,TString,bool)
            """

        @property
        def browse_mode(self) -> Any:
            """
            Return true if HotSwap mode is active for any target.
            """

        @property
        def canonical_parent(self) -> Any:
            """
            Get the canonical parent of the application view.
            """

        @property
        def focused_document_view(self) -> Any:
            """
            Return the name of the document view ('Session' or 'Arranger')
            shown in the currently selected window.
            """

    def add_average_process_usage_listener(self, listener: Callable) -> None:
        """
        add_average_process_usage_listener( (Application)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "average_process_usage" has changed.

            C++ signature :
                void add_average_process_usage_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def add_control_surfaces_listener(self, listener: Callable) -> None:
        """
        add_control_surfaces_listener( (Application)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "control_surfaces" has changed.

            C++ signature :
                void add_control_surfaces_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def add_open_dialog_count_listener(self, listener: Callable) -> None:
        """
        add_open_dialog_count_listener( (Application)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "open_dialog_count" has changed.

            C++ signature :
                void add_open_dialog_count_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def add_peak_process_usage_listener(self, listener: Callable) -> None:
        """
        add_peak_process_usage_listener( (Application)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "peak_process_usage" has changed.

            C++ signature :
                void add_peak_process_usage_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def add_unavailable_features_listener(self, listener: Callable) -> None:
        """
        add_unavailable_features_listener( (Application)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "unavailable_features" has changed.

            C++ signature :
                void add_unavailable_features_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def average_process_usage_has_listener(self, listener: Callable) -> bool:
        """
        average_process_usage_has_listener( (Application)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "average_process_usage".

            C++ signature :
                bool average_process_usage_has_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def control_surfaces_has_listener(self, listener: Callable) -> bool:
        """
        control_surfaces_has_listener( (Application)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "control_surfaces".

            C++ signature :
                bool control_surfaces_has_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def get_bugfix_version(self) -> int:
        """
        get_bugfix_version( (Application)arg1) -> int :
            Returns an integer representing the bugfix version of Live.

            C++ signature :
                int get_bugfix_version(TPyHandle<ASongApp>)
        """

    def get_build_id(self) -> str:
        """
        get_build_id( (Application)arg1) -> str :
            Returns a string identifying the build.

            C++ signature :
                TString get_build_id(TPyHandle<ASongApp>)
        """

    def get_document(self) -> Song:
        """
        get_document( (Application)arg1) -> Song :
            Returns the current Live Set.

            C++ signature :
                TWeakPtr<TPyHandle<ASong>> get_document(TPyHandle<ASongApp>)
        """

    def get_major_version(self) -> int:
        """
        get_major_version( (Application)arg1) -> int :
            Returns an integer representing the major version of Live.

            C++ signature :
                int get_major_version(TPyHandle<ASongApp>)
        """

    def get_minor_version(self) -> int:
        """
        get_minor_version( (Application)arg1) -> int :
            Returns an integer representing the minor version of Live.

            C++ signature :
                int get_minor_version(TPyHandle<ASongApp>)
        """

    def get_variant(self) -> str:
        """
        get_variant( (Application)arg1) -> str :
            Returns one of the strings in Live.Application.Variants.

            C++ signature :
                TString get_variant(TPyHandle<ASongApp>)
        """

    def get_version_string(self) -> str:
        """
        get_version_string( (Application)arg1) -> str :
            Returns the full version string of Live.

            C++ signature :
                TString get_version_string(TPyHandle<ASongApp>)
        """

    def has_option(self, arg2: object) -> bool:
        """
        has_option( (Application)arg1, (object)arg2) -> bool :
            Returns True if the given entry exists in Options.txt, False otherwise.

            C++ signature :
                bool has_option(TPyHandle<ASongApp>,TString)
        """

    def open_dialog_count_has_listener(self, listener: Callable) -> bool:
        """
        open_dialog_count_has_listener( (Application)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "open_dialog_count".

            C++ signature :
                bool open_dialog_count_has_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def peak_process_usage_has_listener(self, listener: Callable) -> bool:
        """
        peak_process_usage_has_listener( (Application)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "peak_process_usage".

            C++ signature :
                bool peak_process_usage_has_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def press_current_dialog_button(self, arg2: int) -> None:
        """
        press_current_dialog_button( (Application)arg1, (int)arg2) -> None :
            Press a button, by index, on the current message box.

            C++ signature :
                void press_current_dialog_button(TPyHandle<ASongApp>,int)
        """

    def remove_average_process_usage_listener(self, listener: Callable) -> None:
        """
        remove_average_process_usage_listener( (Application)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "average_process_usage".

            C++ signature :
                void remove_average_process_usage_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def remove_control_surfaces_listener(self, listener: Callable) -> None:
        """
        remove_control_surfaces_listener( (Application)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "control_surfaces".

            C++ signature :
                void remove_control_surfaces_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def remove_open_dialog_count_listener(self, listener: Callable) -> None:
        """
        remove_open_dialog_count_listener( (Application)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "open_dialog_count".

            C++ signature :
                void remove_open_dialog_count_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def remove_peak_process_usage_listener(self, listener: Callable) -> None:
        """
        remove_peak_process_usage_listener( (Application)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "peak_process_usage".

            C++ signature :
                void remove_peak_process_usage_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def remove_unavailable_features_listener(self, listener: Callable) -> None:
        """
        remove_unavailable_features_listener( (Application)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "unavailable_features".

            C++ signature :
                void remove_unavailable_features_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    def show_message(
        self, text: object, buttons: int = 0, enable_markup: bool = False, show_success_icon: bool = False
    ) -> int:
        """
        show_message( (Application)arg1, (Text)text [, (int)buttons=Application.MessageButtons.OK_BUTTON [, (bool)enable_markup=False [, (bool)show_success_icon=False]]]) -> int :
            Shows a message box, returning the position of the pressed button.

            C++ signature :
                int show_message(TPyHandle<ASongApp>,TText [,int=Application.MessageButtons.OK_BUTTON [,bool=False [,bool=False]]])
        """

    def show_on_the_fly_message(
        self, message: str, buttons: int = 0, enable_markup: bool = False, show_success_icon: bool = False
    ) -> int:
        """
        show_on_the_fly_message( (Application)arg1, (str)message [, (int)buttons=Application.MessageButtons.OK_BUTTON [, (bool)enable_markup=False [, (bool)show_success_icon=False]]]) -> int :
            Same as show_message, but for when there is no predefined Text object.

            C++ signature :
                int show_on_the_fly_message(TPyHandle<ASongApp>,std::__1::basic_string<char, std::__1::char_traits<char>, std::__1::allocator<char>> [,int=Application.MessageButtons.OK_BUTTON [,bool=False [,bool=False]]])
        """

    def unavailable_features_has_listener(self, listener: Callable) -> bool:
        """
        unavailable_features_has_listener( (Application)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "unavailable_features".

            C++ signature :
                bool unavailable_features_has_listener(TPyHandle<ASongApp>,boost::python::api::object)
        """

    @property
    def average_process_usage(self) -> Any:
        """
        Reports Live's average CPU load.
        """

    @property
    def browser(self) -> Any:
        """
        Returns an interface to the browser.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Returns the canonical parent of the application.
        """

    @property
    def control_surfaces(self) -> Any:
        """
        Const access to a list of the control surfaces selected in preferences, in the same order.
        The list contains None if no control surface is active at that index.
        """

    @property
    def current_dialog_button_count(self) -> Any:
        """
        Number of buttons on the current dialog.
        """

    @property
    def current_dialog_message(self) -> Any:
        """
        Text of the last dialog that appeared; Empty if all dialogs just disappeared.
        """

    @property
    def number_of_push_apps_running(self) -> Any:
        """
        Returns the number of connected Push apps.
        """

    @property
    def open_dialog_count(self) -> Any:
        """
        The number of open dialogs in Live. 0 if not dialog is open.
        """

    @property
    def peak_process_usage(self) -> Any:
        """
        Reports Live's peak CPU load.
        """

    @property
    def unavailable_features(self) -> Any:
        """
        List of features that are unavailable due to limitations of the current Live edition.
        """

    @property
    def view(self) -> View:
        """
        Returns the applications view component.
        """


class ControlDescription:
    """
    Describes a control present in a control surface proxy
    """

    @property
    def id(self) -> Any:
        pass

    @property
    def name(self) -> str:
        pass


class ControlDescriptionVector:
    """
    A container for returning control descriptions.
    """

    def append(self, arg2: object) -> None:
        """
        append( (ControlDescriptionVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<TControlDescription, std::__1::allocator<TControlDescription>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (ControlDescriptionVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<TControlDescription, std::__1::allocator<TControlDescription>> {lvalue},boost::python::api::object)
        """


class ControlSurfaceProxy:
    """
    Represents a control surface running in a different process. For use by M4L
    """

    def add_control_values_arrived_listener(self, listener: Callable) -> None:
        """
        add_control_values_arrived_listener( (ControlSurfaceProxy)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "control_values_arrived" has changed.

            C++ signature :
                void add_control_values_arrived_listener(APythonControlSurfaceProxy,boost::python::api::object)
        """

    def add_midi_received_listener(self, listener: Callable) -> None:
        """
        add_midi_received_listener( (ControlSurfaceProxy)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "midi_received" has changed.

            C++ signature :
                void add_midi_received_listener(APythonControlSurfaceProxy,boost::python::api::object)
        """

    def control_values_arrived_has_listener(self, listener: Callable) -> bool:
        """
        control_values_arrived_has_listener( (ControlSurfaceProxy)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "control_values_arrived".

            C++ signature :
                bool control_values_arrived_has_listener(APythonControlSurfaceProxy,boost::python::api::object)
        """

    def enable_receive_midi(self, arg2: bool) -> None:
        """
        enable_receive_midi( (ControlSurfaceProxy)arg1, (bool)arg2) -> None :

            C++ signature :
                void enable_receive_midi(APythonControlSurfaceProxy {lvalue},bool)
        """

    def fetch_received_midi_messages(self) -> tuple:
        """
        fetch_received_midi_messages( (ControlSurfaceProxy)arg1) -> tuple :

            C++ signature :
                boost::python::tuple fetch_received_midi_messages(APythonControlSurfaceProxy {lvalue})
        """

    def fetch_received_values(self) -> tuple:
        """
        fetch_received_values( (ControlSurfaceProxy)arg1) -> tuple :

            C++ signature :
                boost::python::tuple fetch_received_values(APythonControlSurfaceProxy {lvalue})
        """

    def grab_control(self, arg2: int) -> None:
        """
        grab_control( (ControlSurfaceProxy)arg1, (int)arg2) -> None :

            C++ signature :
                void grab_control(APythonControlSurfaceProxy {lvalue},int)
        """

    def midi_received_has_listener(self, listener: Callable) -> bool:
        """
        midi_received_has_listener( (ControlSurfaceProxy)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "midi_received".

            C++ signature :
                bool midi_received_has_listener(APythonControlSurfaceProxy,boost::python::api::object)
        """

    def release_control(self, arg2: int) -> None:
        """
        release_control( (ControlSurfaceProxy)arg1, (int)arg2) -> None :

            C++ signature :
                void release_control(APythonControlSurfaceProxy {lvalue},int)
        """

    def remove_control_values_arrived_listener(self, listener: Callable) -> None:
        """
        remove_control_values_arrived_listener( (ControlSurfaceProxy)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "control_values_arrived".

            C++ signature :
                void remove_control_values_arrived_listener(APythonControlSurfaceProxy,boost::python::api::object)
        """

    def remove_midi_received_listener(self, listener: Callable) -> None:
        """
        remove_midi_received_listener( (ControlSurfaceProxy)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "midi_received".

            C++ signature :
                void remove_midi_received_listener(APythonControlSurfaceProxy,boost::python::api::object)
        """

    def send_midi(self, arg2: object) -> None:
        """
        send_midi( (ControlSurfaceProxy)arg1, (tuple)arg2) -> None :

            C++ signature :
                void send_midi(APythonControlSurfaceProxy {lvalue},boost::python::tuple)
        """

    def send_value(self, arg2: object) -> None:
        """
        send_value( (ControlSurfaceProxy)arg1, (tuple)arg2) -> None :

            C++ signature :
                void send_value(APythonControlSurfaceProxy {lvalue},boost::python::tuple)
        """

    def subscribe_to_control(self, arg2: int) -> None:
        """
        subscribe_to_control( (ControlSurfaceProxy)arg1, (int)arg2) -> None :

            C++ signature :
                void subscribe_to_control(APythonControlSurfaceProxy {lvalue},int)
        """

    def unsubscribe_from_control(self, arg2: int) -> None:
        """
        unsubscribe_from_control( (ControlSurfaceProxy)arg1, (int)arg2) -> None :

            C++ signature :
                void unsubscribe_from_control(APythonControlSurfaceProxy {lvalue},int)
        """

    @property
    def control_descriptions(self) -> Any:
        pass

    @property
    def type_name(self) -> str:
        pass


class MessageButtons:
    """
    Class that represent an enumeration of values for MessageButtons
    Specifies the characteristics of the message box, e.g. which buttons to show.
    """

    OK_BUTTON = 0


class UnavailableFeature:
    """
    Class that represent an enumeration of values for UnavailableFeature
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    note_velocity_ranges_and_probabilities = 0


class UnavailableFeatureVector:
    """
    A container for returning unavailable features.
    """

    def append(self, arg2: object) -> None:
        """
        append( (UnavailableFeatureVector)arg1, (object)arg2) -> None :

            C++ signature :
                void append(std::__1::vector<NPythonApplication::TUnavailableFeature, std::__1::allocator<NPythonApplication::TUnavailableFeature>> {lvalue},boost::python::api::object)
        """

    def extend(self, arg2: object) -> None:
        """
        extend( (UnavailableFeatureVector)arg1, (object)arg2) -> None :

            C++ signature :
                void extend(std::__1::vector<NPythonApplication::TUnavailableFeature, std::__1::allocator<NPythonApplication::TUnavailableFeature>> {lvalue},boost::python::api::object)
        """


class Variants:
    """
    Holds strings representing what type of Live is running.
    """

    BETA = "Beta"
    INTRO = "Intro"
    LITE = "Lite"
    STANDARD = "Standard"
    SUITE = "Suite"
    TRIAL = "Trial"


def combine_apcs() -> bool:
    """
    combine_apcs() -> bool :
        Returns true if multiple APCs should be combined.

        C++ signature :
            bool combine_apcs()
    """


def encrypt_challenge(dongle1: int, dongle2: int, key_index: int = 0) -> tuple:
    """
    encrypt_challenge( (int)dongle1, (int)dongle2 [, (int)key_index=0]) -> tuple :
        Returns an encrypted challenge based on the TEA algortithm

        C++ signature :
            boost::python::tuple encrypt_challenge(int,int [,int=0])
    """


def encrypt_challenge2(arg1: int) -> int:
    """
    encrypt_challenge2( (int)arg1) -> int :
        Returns the UMAC hash for the given challenge.

        C++ signature :
            int encrypt_challenge2(int)
    """


def get_application() -> Application:
    """
    get_application() -> Application :
        Returns the application instance.

        C++ signature :
            TWeakPtr<TPyHandle<ASongApp>> get_application()
    """


def get_random_int(arg1: int, arg2: int) -> int:
    """
    get_random_int( (int)arg1, (int)arg2) -> int :
        Returns a random integer from the given range.

        C++ signature :
            int get_random_int(int,int)
    """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/base/proxy.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3257 bytes
from .util import BooleanContext, old_hasattr


class ProxyBase:
    _skip_wrapper_lookup = None

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._skip_wrapper_lookup = BooleanContext()

    def proxy_old_hasattr(self, attr):
        with self._skip_wrapper_lookup():
            return old_hasattr(self, attr)

    def __getattr__(self, name):
        if not self._skip_wrapper_lookup:
            obj = self.proxied_object
            interface = self.proxied_interface
            if obj:
                if old_hasattr(interface, name):
                    return getattr(obj, name)
            if interface:
                if old_hasattr(interface, name):
                    return getattr(interface, name)
        raise AttributeError(f"Does not have attribute {name}")

    def __setattr__(self, name, value):
        obj = self.proxied_object
        interface = self.proxied_interface
        if obj and old_hasattr(interface, name):
            if self.proxy_old_hasattr(name):
                raise AttributeError(f"Ambiguous set attribute: {name} proxied: {obj}")
            setattr(obj, name, value)
        else:
            if old_hasattr(interface, name):
                raise AttributeError(f"Ambiguous set attribute: {name} proxied: {obj}")
            self.__dict__[name] = value

    @property
    def proxied_object(self):
        pass

    @property
    def proxied_interface(self):
        obj = self.proxied_object
        return getattr(obj, "proxied_interface", obj)


class Proxy(ProxyBase):
    proxied_object = None
    _proxied_interface = None

    @property
    def proxied_interface(self):
        return self._proxied_interface or super().proxied_interface

    def __init__(self, proxied_object=None, proxied_interface=None, *a, **k):
        (super().__init__)(*a, **k)
        self.proxied_object = proxied_object
        self._proxied_interface = proxied_interface



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/logical_display_segment.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2461 bytes


class LogicalDisplaySegment:
    separator = ""

    def __init__(self, width=None, update_callback=None, *a, **k):
        (super().__init__)(*a, **k)
        self._update_callback = update_callback
        self._width = width
        self._position_identifier = ()
        self._data_source = None
        self._display_string = None

    def disconnect(self):
        self._update_callback = None
        self._position_identifier = None
        if self._data_source is not None:
            self._data_source.set_update_callback(None)
            self._data_source = None

    def set_data_source(self, data_source):
        if self._data_source is not None:
            self._data_source.set_update_callback(None)
        self._data_source = data_source
        if self._data_source is not None:
            self._data_source.set_update_callback(self.update)
        self._display_string = self._get_display_string()

    def data_source(self):
        return self._data_source

    def set_position_identifier(self, position_identifier):
        self._position_identifier = position_identifier

    def position_identifier(self):
        return self._position_identifier

    def update(self):
        if self._update_callback:
            self._display_string = self._get_display_string()
            self._update_callback()

    def _get_display_string(self):
        if self._data_source is not None:
            separator = self._data_source.separator + self.separator
            width = self._width - len(separator)
            return self._data_source.adjust_string(width) + separator
        return " " * int(self._width)

    def display_string(self):
        if self._display_string is None:
            self._display_string = self._get_display_string()
        return self._display_string

    def __str__(self):
        return self.display_string()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/control_surface.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 20413 bytes
from contextlib import contextmanager, nullcontext

from display.view.util import suppress_notifications

from ableton.v2.control_surface import SimpleControlSurface

from ..base import const, find_if, inject, lazy_attribute, listens
from ..live import is_song_recording, is_track_armed, liveobj_valid
from . import DeviceBankRegistry
from .components import (
    BackgroundComponent,
    DrumGroupComponent,
    SessionComponent,
    SlicedSimplerComponent,
    ViewControlComponent,
)
from .consts import M4L_PRIORITY
from .control_surface_mapping import ControlSurfaceMappingMixin
from .display import Display, Renderable
from .identification import IdentificationComponent
from .instrument_finder import InstrumentFinderComponent
from .layer import Layer
from .session_ring_selection_linking import SessionRingSelectionLinking


def create_control_surface(name=None, specification=None, c_instance=None):
    return type(name, (ControlSurface,), {})(specification, c_instance=c_instance)


class ControlSurface(SimpleControlSurface, ControlSurfaceMappingMixin):
    def __init__(self, specification=None, *a, **k):
        (super().__init__)(a, specification=specification, **k)
        is_identifiable = (
            specification.identity_response_id_bytes is not None or specification.custom_identity_response is not None
        )
        self.specification = specification
        self.device_provider = None
        self.device_bank_registry = None
        self.display = None
        self.instrument_finder = None
        self.session_ring_selection_linking = None
        self._can_update_controlled_track = False
        self._can_enable_auto_arm = False
        self._deferring_render_and_update_display = nullcontext
        self._drum_group_component = None
        self._sliced_simpler_component = None
        self._control_surface_injector.update(
            react=(const(self.react)),
            notifications=(const(specification.display_specification.notifications)),
            suppress_notifications=(lambda: self._suppress_notifications),
        )
        with inject(skin=(const(specification.control_surface_skin))).everywhere():
            with self._control_surface_injector:
                self.elements = self._create_elements(specification)
                self._element_injector = inject(element_container=(const(self.elements))).everywhere()
        with self.component_guard():
            self._on_entering_component_guard()
            self._background = self._create_background(specification.background_priority)
            self._identification = self._create_identification(specification) if is_identifiable else None
            self._target_track = specification.target_track_component_type()
            self._ControlSurface__on_target_track_changed.subject = self._target_track
            self._create_feedback_related_listeners()
            with self._create_extended_injector():
                if is_identifiable:
                    self._auto_arm = (
                        specification.auto_arm_component_type(is_enabled=False)
                        if specification.include_auto_arming
                        else None
                    )
                    self.setup()
                    if self._drum_group_component or self._sliced_simpler_component:
                        self._ControlSurface__on_instrument_changed()
                    if specification.display_specification.protocol:
                        self.display = self.register_disconnectable(
                            Display(specification.display_specification, self.renderable_components, self.elements),
                        )
                        self._deferring_render_and_update_display = self.display.deferring_render_and_update_display
        self._can_enable_session_ring = (
            is_identifiable and find_if(lambda x: isinstance(x, SessionComponent), self._components) is not None
        )

    def disconnect(self):
        if self.specification.send_goodbye_messages_last:
            super().disconnect()
            self._send_specification_messages(messages_name="goodbye_messages")
        else:
            self._send_specification_messages(messages_name="goodbye_messages")
            super().disconnect()
        self.specification = None
        self.elements = None
        self.device_provider = None
        self.device_bank_registry = None
        self.instrument_finder = None
        self.session_ring_selection_linking = None
        self._drum_group_component = None
        self._sliced_simpler_component = None

    @property
    def renderable_components(self):
        return tuple(comp for comp in self._components if isinstance(comp, Renderable))

    @property
    def controls(self):
        return [control for control in self._controls if not control.is_private]

    def port_settings_changed(self):
        if self._identification:
            self._identification.request_identity()
        else:
            self._send_specification_messages()
            self.refresh_state()

    def on_identified(self, response_bytes):
        self._send_specification_messages()
        self.refresh_state()

    def refresh_state(self):
        super().refresh_state()
        if self.specification.feedback_channels is not None:
            self._c_instance.set_feedback_channels(self.specification.feedback_channels)
        self._ControlSurface__on_target_track_changed()
        self._update_feedback_velocity()

    def can_lock_to_devices(self):
        return self.device_provider is not None

    def lock_to_device(self, device):
        with self.component_guard():
            self.device_provider.lock_to_device(device)

    def unlock_from_device(self, _):
        with self.component_guard():
            self.device_provider.unlock_from_device()

    def restore_bank(self, bank_index):
        device = self.device_provider.device
        if self.device_provider.is_locked_to_device:
            if liveobj_valid(device):
                with self.component_guard():
                    self.device_bank_registry.set_device_bank(device, bank_index)

    def target_track_changed(self, target_track):
        pass

    def drum_group_changed(self, drum_group):
        pass

    def sliced_simpler_changed(self, sliced_simpler):
        pass

    def set_can_update_controlled_track(self, can_update):
        self._can_update_controlled_track = can_update
        self._update_controlled_track()

    def set_can_auto_arm(self, can_auto_arm):
        self._can_enable_auto_arm = can_auto_arm
        self._update_auto_arm()

    def update(self):
        super().update()
        if self.display:
            self.display.clear_content_cache()

    def update_display(self):
        super().update_display()
        self._render_and_update_display()

    def react(self, event):
        if self.display:
            self.display.react(event)

    @staticmethod
    def mxd_grab_control_priority():
        return M4L_PRIORITY

    def _render_and_update_display(self):
        if self.display:
            with self._control_surface_injector:
                self.display.render_and_update_display()

    def _suppress_notifications(self):
        if self.display:
            suppress_notifications(self.display.state)

    @staticmethod
    def _should_include_element_in_background(element):
        return True

    def _get_additional_dependencies(self):
        pass

    def _on_entering_component_guard(self):
        pass

    def _send_specification_messages(self, messages_name="hello_messages"):
        for msg in getattr(self.specification, messages_name) or []:
            self._send_midi(msg)

    @contextmanager
    def _component_guard(self):
        with super()._component_guard():
            with self._element_injector:
                with self._deferring_render_and_update_display():
                    yield

    @staticmethod
    def _create_elements(specification):
        return specification.elements_type()

    def _create_background(self, priority):
        layer_dict = {c.name: c for c in [] if self._should_include_element_in_background(c)}
        layer_dict["priority"] = priority
        background = BackgroundComponent(is_enabled=False, layer=Layer(**layer_dict))
        background.set_enabled(True)
        return background

    def _create_identification(self, specification):
        identification = IdentificationComponent(
            identity_request=(specification.identity_request),
            identity_request_delay=(specification.identity_request_delay),
            identity_response_id_bytes=(specification.identity_response_id_bytes),
            custom_identity_response=(specification.custom_identity_response),
        )
        self._ControlSurface__on_is_identified_changed.subject = identification
        return identification

    def _create_instrument_finder(self):
        if self.instrument_finder is None:
            self.instrument_finder = InstrumentFinderComponent()
            self._ControlSurface__on_instrument_changed.subject = self.instrument_finder
        return self.instrument_finder

    @lazy_attribute
    def _create_session_ring(self):
        self._session_ring = self.specification.session_ring_component_type(
            is_enabled=False,
            num_tracks=(self.specification.num_tracks),
            num_scenes=(self.specification.num_scenes),
            include_returns=(self.specification.include_returns),
            include_master=(self.specification.include_master),
            right_align_non_player_tracks=(self.specification.right_align_non_player_tracks),
            snap_track_offset=(self.specification.snap_track_offset),
        )
        return self._session_ring

    @lazy_attribute
    def _create_device_provider(self):
        self.device_provider = self.register_disconnectable(self.specification.device_provider_type(song=(self.song)))
        self.device_provider.update_device_selection()
        return self.device_provider

    @lazy_attribute
    def _create_device_bank_registry(self):
        self.device_bank_registry = self.register_disconnectable(DeviceBankRegistry())
        return self.device_bank_registry

    def _create_extended_injector(self):
        inject_dict = {
            "full_velocity": const(self._c_instance.full_velocity),
            "target_track": const(self._target_track),
            "session_ring": lambda: self._create_session_ring,
            "device_provider": lambda: self._create_device_provider,
            "device_bank_registry": lambda: self._create_device_bank_registry,
            "toggle_lock": const(self._c_instance.toggle_lock),
            "playhead": const(self._c_instance.playhead),
        }
        inject_dict.update(self._get_additional_dependencies() or {})
        return inject(**inject_dict).everywhere()

    def _create_feedback_related_listeners(self):
        self.register_slot(self.song, self._update_feedback_velocity, "session_record")
        self.register_slot(self.song, self._update_feedback_velocity, "record_mode")

    def _register_component(self, component):
        super()._register_component(component)
        if isinstance(component, DrumGroupComponent):
            self._drum_group_component = component
            self._create_instrument_finder()
        if isinstance(component, SlicedSimplerComponent):
            self._sliced_simpler_component = component
            self._create_instrument_finder()
        if isinstance(component, ViewControlComponent):
            link_to_tracks = self.specification.link_session_ring_to_track_selection
            link_to_scenes = self.specification.link_session_ring_to_scene_selection
            if link_to_tracks or link_to_scenes:
                self.session_ring_selection_linking = self.register_disconnectable(
                    SessionRingSelectionLinking(
                        selection_changed_notifier=component,
                        link_to_track_selection=link_to_tracks,
                        link_to_scene_selection=link_to_scenes,
                    ),
                )

    def _update_feedback_velocity(self):
        track = self._target_track.target_track
        if is_song_recording() and is_track_armed(track):
            feedback_velocity = self.specification.recording_feedback_velocity
        else:
            feedback_velocity = self.specification.playing_feedback_velocity
        self._c_instance.set_feedback_velocity(int(feedback_velocity))
        self._c_instance.playhead.velocity = int(feedback_velocity)

    def _update_controlled_track(self):
        if self._can_update_controlled_track:
            self.set_controlled_track(self._target_track.target_track)
        else:
            self.release_controlled_track()

    def _update_auto_arm(self):
        if self._auto_arm:
            self._auto_arm.set_enabled(self._identification.is_identified and self._can_enable_auto_arm)

    def _update_instrument_components(self):
        self._update_controlled_track()
        if self._drum_group_component:
            drum_group = self.instrument_finder.drum_group
            self._drum_group_component.set_drum_group_device(drum_group)
            self.drum_group_changed(drum_group)
        if self._sliced_simpler_component:
            sliced_simpler = self.instrument_finder.sliced_simpler
            self._sliced_simpler_component.set_simpler_device(sliced_simpler)
            self.sliced_simpler_changed(sliced_simpler)

    @listens("target_track")
    def __on_target_track_changed(self):
        track = self._target_track.target_track
        self._ControlSurface__on_track_arm_changed.subject = track
        self._ControlSurface__on_track_implicit_arm_changed.subject = track
        self._update_controlled_track()
        self.target_track_changed(track)

    @listens("arm")
    def __on_track_arm_changed(self):
        self._update_feedback_velocity()

    @listens("implicit_arm")
    def __on_track_implicit_arm_changed(self):
        self._update_feedback_velocity()

    @listens("instrument")
    def __on_instrument_changed(self):
        self._update_instrument_components()

    @listens("is_identified")
    def __on_is_identified_changed(self, is_identified):
        if is_identified:
            self.on_identified(self._identification.received_response_bytes)
        if self._can_enable_session_ring:
            self._session_ring.set_enabled(is_identified)
        self._update_auto_arm()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/DeviceComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 19268 bytes
import Live
from _Generic.Devices import (
    best_of_parameter_bank,
    device_parameters_to_map,
    number_of_parameter_banks,
    parameter_bank_names,
    parameter_banks,
)
from ableton.v2.base import liveobj_valid

from .ControlSurfaceComponent import ControlSurfaceComponent
from .DeviceBankRegistry import DeviceBankRegistry
from .DisplayDataSource import DisplayDataSource
from .SubjectSlot import Subject, subject_slot, subject_slot_group


def device_to_appoint(device):
    appointed_device = device
    if not liveobj_valid(device) or device.can_have_drum_pads:
        if not device.has_macro_mappings:
            if len(device.chains) > 0:
                if liveobj_valid(device.view.selected_chain):
                    if len(device.view.selected_chain.devices) > 0:
                        appointed_device = device_to_appoint(device.view.selected_chain.devices[0])
        return appointed_device
    return None


def select_and_appoint_device(song, device_to_select, ignore_unmapped_macros=True):
    appointed_device = device_to_select
    if ignore_unmapped_macros:
        appointed_device = device_to_appoint(device_to_select)
    song.appointed_device = appointed_device
    song.view.select_device(device_to_select, False)


class DeviceComponent(ControlSurfaceComponent, Subject):
    __subject_events__ = ("device",)

    def __init__(self, device_bank_registry=None, device_selection_follows_track_selection=False, *a, **k):
        (super().__init__)(*a, **k)
        self._device_bank_registry = device_bank_registry or DeviceBankRegistry()
        self._device = None
        self._parameter_controls = None
        self._bank_up_button = None
        self._bank_down_button = None
        self._bank_buttons = None
        self._on_off_button = None
        self._lock_button = None
        self._lock_callback = None
        self._device_name_data_source = None
        self._bank_index = 0
        self._bank_name = "<No Bank>"
        self._locked_to_device = False

        def make_property_slot(name, alias=None):
            alias = alias or name
            return self.register_slot(None, getattr(self, f"_on_{alias}_changed"), name)

        self._on_off_property_slot = make_property_slot("value", alias="device_on_off")
        self._name_property_slot = make_property_slot("name", alias="device_name")
        self._parameters_property_slot = make_property_slot("parameters")
        self._device_bank_property_slot = make_property_slot("device_bank")

        def make_button_slot(name):
            return self.register_slot(None, getattr(self, f"_{name}_value"), "value")

        self._bank_up_button_slot = make_button_slot("bank_up")
        self._bank_down_button_slot = make_button_slot("bank_down")
        self._lock_button_slot = make_button_slot("lock")
        self._on_off_button_slot = make_button_slot("on_off")
        song = self.song()
        view = song.view
        self.device_selection_follows_track_selection = device_selection_follows_track_selection
        self._device_bank_property_slot.subject = self._device_bank_registry
        self._DeviceComponent__on_appointed_device_changed.subject = song
        self._DeviceComponent__on_selected_track_changed.subject = view
        self._DeviceComponent__on_selected_device_changed.subject = view.selected_track.view

    def disconnect(self):
        self._device_bank_registry = None
        self._lock_callback = None
        self._release_parameters(self._parameter_controls)
        self._parameter_controls = None
        self._bank_up_button = None
        self._bank_down_button = None
        self._bank_buttons = None
        self._on_off_button = None
        self._lock_button = None
        self._device = None
        super().disconnect()

    def on_enabled_changed(self):
        self.update()

    def device(self):
        return self._device

    @subject_slot("appointed_device")
    def __on_appointed_device_changed(self):
        self.set_device(device_to_appoint(self.song().appointed_device))

    def set_device(self, device):
        if self._locked_to_device or device != self._device or type(device) != type(self._device):
            if liveobj_valid(self._device):
                self._release_parameters(self._parameter_controls)
            self._device = device
            self._name_property_slot.subject = device
            self._parameters_property_slot.subject = device
            self._on_off_property_slot.subject = self._on_off_parameter()
            if liveobj_valid(self._device):
                self._bank_index = 0
            self._bank_index = self._device_bank_registry.get_device_bank(self._device)
            self._bank_name = "<No Bank>"
            self._on_device_name_changed()
            self.update()
            self.notify_device()

    @subject_slot("has_macro_mappings")
    def __on_has_macro_mappings_changed(self):
        self.song().appointed_device = device_to_appoint(self._DeviceComponent__on_has_macro_mappings_changed.subject)

    @subject_slot("selected_track")
    def __on_selected_track_changed(self):
        self._DeviceComponent__on_selected_device_changed.subject = self.song().view.selected_track.view
        if self.device_selection_follows_track_selection:
            self.update_device_selection()

    @subject_slot("chains")
    def __on_chains_changed(self):
        self._update_appointed_device()

    @subject_slot("selected_device")
    def __on_selected_device_changed(self):
        self._update_appointed_device()

    def _update_appointed_device(self):
        song = self.song()
        device = song.view.selected_track.view.selected_device
        if liveobj_valid(device):
            song.appointed_device = device_to_appoint(device)
        rack_device = device if isinstance(device, Live.RackDevice.RackDevice) else None
        self._DeviceComponent__on_has_macro_mappings_changed.subject = rack_device
        self._DeviceComponent__on_chains_changed.subject = rack_device

    def update_device_selection(self):
        view = self.song().view
        track_or_chain = view.selected_chain if view.selected_chain else view.selected_track
        device_to_select = None
        if isinstance(track_or_chain, Live.Track.Track):
            device_to_select = track_or_chain.view.selected_device
        if not liveobj_valid(device_to_select):
            if len(track_or_chain.devices) > 0:
                device_to_select = track_or_chain.devices[0]
        if liveobj_valid(device_to_select):
            appointed_device = device_to_appoint(device_to_select)
            view.select_device(device_to_select, False)
            self.song().appointed_device = appointed_device
            self.set_device(appointed_device)
        else:
            self.song().appointed_device = None
            self.set_device(None)

    def set_bank_prev_button(self, button):
        if button != self._bank_down_button:
            self._bank_down_button = button
            self._bank_down_button_slot.subject = button
            self.update()

    def set_bank_next_button(self, button):
        if button != self._bank_up_button:
            self._bank_up_button = button
            self._bank_up_button_slot.subject = button
            self.update()

    def set_bank_nav_buttons(self, down_button, up_button):
        self.set_bank_prev_button(down_button)
        self.set_bank_next_button(up_button)

    def set_bank_buttons(self, buttons):
        self._bank_buttons = buttons
        self._on_bank_value.replace_subjects(buttons or [])
        self.update()

    def set_parameter_controls(self, controls):
        self._release_parameters(self._parameter_controls)
        self._parameter_controls = controls
        self.update()

    def set_lock_to_device(self, lock, device):
        if lock:
            self.set_device(device)
        self._locked_to_device = lock
        self._update_lock_button()

    def set_lock_button(self, button):
        self._lock_button = button
        self._lock_button_slot.subject = button
        self._update_lock_button()

    def set_on_off_button(self, button):
        self._on_off_button = button
        self._on_off_button_slot.subject = button
        self._update_on_off_button()

    def set_lock_callback(self, callback):
        self._lock_callback = callback

    def restore_bank(self, bank_index):
        if liveobj_valid(self._device):
            if self._is_banking_enabled():
                if self._locked_to_device:
                    if self._number_of_parameter_banks() > bank_index:
                        if self._bank_index != bank_index:
                            self._bank_index = bank_index
                            self.update()

    def device_name_data_source(self):
        if self._device_name_data_source is None:
            self._device_name_data_source = DisplayDataSource()
            self._on_device_name_changed()
        return self._device_name_data_source

    def update(self):
        super().update()
        if self.is_enabled() and liveobj_valid(self._device):
            self._device_bank_registry.set_device_bank(self._device, self._bank_index)
            if self._parameter_controls is not None:
                old_bank_name = self._bank_name
                self._assign_parameters()
                if self._bank_name != old_bank_name:
                    self._show_msg_callback(self._device.name + " Bank: " + self._bank_name)
        else:
            if self._parameter_controls is not None:
                self._release_parameters(self._parameter_controls)
        if self.is_enabled():
            self._update_on_off_button()
            self._update_lock_button()
            self._update_device_bank_buttons()
            self._update_device_bank_nav_buttons()

    def _bank_up_value(self, value):
        if self.is_enabled():
            if not self._bank_up_button.is_momentary() or value != 0:
                if liveobj_valid(self._device):
                    num_banks = self._number_of_parameter_banks()
                    if self._bank_down_button is None:
                        self._bank_name = ""
                        self._bank_index = (self._bank_index + 1) % num_banks if self._bank_index is not None else 0
                        self.update()
        else:
            pass
        if self._bank_index is None or num_banks > self._bank_index + 1:
            self._bank_name = ""
            self._bank_index = self._bank_index + 1 if self._bank_index is not None else 0
            self.update()

    def _bank_down_value(self, value):
        if self.is_enabled():
            if not self._bank_down_button.is_momentary() or value != 0:
                if liveobj_valid(self._device):
                    if self._bank_index is None or self._bank_index > 0:
                        self._bank_name = ""
                        self._bank_index = (
                            self._bank_index - 1
                            if self._bank_index is not None
                            else max(0, self._number_of_parameter_banks() - 1)
                        )
                        self.update()

    def _lock_value(self, value):
        if not self._lock_button.is_momentary() or value != 0:
            self._lock_callback()

    def _on_off_value(self, value):
        if not self._on_off_button.is_momentary() or value != 0:
            parameter = self._on_off_parameter()
            if liveobj_valid(parameter):
                if parameter.is_enabled:
                    parameter.value = float(int(parameter.value == 0.0))

    @subject_slot_group("value")
    def _on_bank_value(self, value, button):
        self._bank_value(value, button)

    def _bank_value(self, value, button):
        if self.is_enabled():
            if liveobj_valid(self._device):
                if not button.is_momentary() or value != 0:
                    bank = list(self._bank_buttons).index(button)
                    if bank != self._bank_index:
                        if self._number_of_parameter_banks() > bank:
                            self._bank_name = ""
                            self._bank_index = bank
                            self.update()
                    else:
                        self._show_msg_callback(self._device.name + " Bank: " + self._bank_name)

    def _is_banking_enabled(self):
        direct_banking = self._bank_buttons is not None
        roundtrip_banking = self._bank_up_button is not None
        increment_banking = self._bank_up_button is not None and self._bank_down_button is not None
        return direct_banking or roundtrip_banking or increment_banking

    def _assign_parameters(self):
        self._bank_name, bank = self._current_bank_details()
        for control, parameter in zip(self._parameter_controls, bank):
            if control is not None:
                if liveobj_valid(parameter):
                    control.connect_to(parameter)
                else:
                    control.release_parameter()

        self._release_parameters(self._parameter_controls[len(bank) :])

    def _on_device_on_off_changed(self):
        self._update_on_off_button()

    def _on_device_name_changed(self):
        if self._device_name_data_source is not None:
            self._device_name_data_source.set_display_string(
                self._device.name if (self.is_enabled()) and (liveobj_valid(self._device)) else "No Device",
            )

    def _on_parameters_changed(self):
        self.update()

    def _on_off_parameter(self):
        result = None
        if liveobj_valid(self._device):
            for parameter in self._device.parameters:
                if str(parameter.name).startswith("Device On"):
                    result = parameter
                    break

        return result

    def _update_on_off_button(self):
        if self.is_enabled():
            if self._on_off_button is not None:
                turn_on = False
                if liveobj_valid(self._device):
                    parameter = self._on_off_parameter()
                    turn_on = liveobj_valid(parameter) and parameter.value > 0.0
                self._on_off_button.set_light(turn_on)

    def _update_lock_button(self):
        if self.is_enabled():
            if self._lock_button is not None:
                self._lock_button.set_light(self._locked_to_device)

    def _update_device_bank_buttons(self):
        if self.is_enabled():
            for index, button in enumerate(self._bank_buttons or []):
                if button:
                    button.set_light(index == self._bank_index and self._device)

    def _update_device_bank_nav_buttons(self):
        if self.is_enabled():
            if self._bank_up_button is not None:
                if self._bank_down_button is not None:
                    can_bank_up = self._bank_index is None or self._number_of_parameter_banks() > self._bank_index + 1
                    can_bank_down = self._bank_index is None or self._bank_index > 0
                    self._bank_up_button.set_light(self._device and can_bank_up)
                    self._bank_down_button.set_light(self._device and can_bank_down)

    def _best_of_parameter_bank(self):
        return best_of_parameter_bank(self._device)

    def _parameter_banks(self):
        return parameter_banks(self._device)

    def _parameter_bank_names(self):
        return parameter_bank_names(self._device)

    def _device_parameters_to_map(self):
        return device_parameters_to_map(self._device)

    def _number_of_parameter_banks(self):
        return number_of_parameter_banks(self._device)

    def _current_bank_details(self):
        bank_name = self._bank_name
        bank = []
        best_of = self._best_of_parameter_bank()
        banks = self._parameter_banks()
        if banks:
            if self._bank_index is not None:
                if self._is_banking_enabled() or not best_of:
                    index = self._bank_index if self._bank_index is not None else 0
                    bank = banks[index]
                    bank_name = self._parameter_bank_names()[index]
                else:
                    bank = best_of
                    bank_name = "Best of Parameters"
        return (bank_name, bank)

    def _on_device_bank_changed(self, device, bank):
        if device == self._device:
            self._bank_index = bank
            self.update()

    def _release_parameters(self, controls):
        if controls is not None:
            for control in controls:
                if control is not None:
                    control.release_parameter()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/controls/toggle_button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1317 bytes
from . import ButtonControl, Connectable, control_event


class ToggleButtonControl(ButtonControl):
    toggled = control_event("toggled")

    class State(ButtonControl.State, Connectable):
        requires_listenable_connected_property = True

        def connect_property(self, *a):
            (super().connect_property)(*a)
            self.is_on = self.connected_property_value

        def on_connected_property_changed(self, value):
            self.is_on = value

        def _on_pressed(self):
            super()._on_pressed()
            self.is_on = not self._is_on
            self._call_listener("toggled", self.is_on)
            self.connected_property_value = self.is_on
